from pathlib import Path
from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class Human(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to text"}
    options = {"output_file": "", "console": True}
    options_desc = {"output_file": "Output to file", "console": "Output to console"}
    emit_graph_trail = False

    def setup(self):
        self.output_file = self.config.get("output_file", "")
        if self.output_file:
            self.output_file = Path(self.output_file)
        else:
            self.output_file = self.scan.home / "output.txt"
        self.helpers.mkdir(self.output_file.parent)
        self._file = None
        return True

    @property
    def file(self):
        if self._file is None:
            self._file = open(self.output_file, mode="a")
        return self._file

    def handle_event(self, event):
        event_type = f"[{event.type}]"
        event_tags = ""
        if getattr(event, "tags", []):
            event_tags = f'\t({", ".join(sorted(getattr(event, "tags", [])))})'
        event_str = f"{event_type:<20}\t{event.data_human}\t{event.module}{event_tags}"
        if self.file is not None:
            self.file.write(event_str + "\n")
            self.file.flush()
        if self.config.get("console", True):
            self.stdout(event_str)

    def cleanup(self):
        if self._file is not None:
            with suppress(Exception):
                self.file.close()



import omegaconf


def flatten_config(config, base="bbot"):
    """
    Flatten a JSON-like config into a list of environment variables:
        {"modules": [{"httpx": {"timeout": 5}}]} --> "BBOT_MODULES_HTTPX_TIMEOUT=5"
    """
    if type(config) == omegaconf.dictconfig.DictConfig:
        for k, v in config.items():
            new_base = f"{base}_{k}"
            if type(v) == omegaconf.dictconfig.DictConfig:
                yield from flatten_config(v, base=new_base)
            elif type(v) != omegaconf.listconfig.ListConfig:
                yield (new_base.upper(), str(v))



import logging
import threading
from pathlib import Path
import concurrent.futures
from omegaconf import OmegaConf
from contextlib import suppress
from collections import OrderedDict

from .stats import ScanStats
from .target import ScanTarget
from .manager import ScanManager
from .dispatcher import Dispatcher
from bbot.modules import module_loader
from bbot.core.event import make_event
from bbot.core.logger import init_logging
from bbot.core.helpers.misc import sha1, rand_string
from bbot.core.helpers.helper import ConfigAwareHelper
from bbot.core.helpers.names_generator import random_name
from bbot.core.helpers.threadpool import ThreadPoolWrapper
from bbot.core.errors import BBOTError, ScanError, ScanCancelledError, ValidationError

log = logging.getLogger("bbot.scanner")

init_logging()


class Scanner:

    _status_codes = {
        "NOT_STARTED": 0,
        "STARTING": 1,
        "RUNNING": 2,
        "FINISHING": 3,
        "CLEANING_UP": 4,
        "ABORTING": 5,
        "ABORTED": 6,
        "FAILED": 7,
        "FINISHED": 8,
    }

    def __init__(
        self,
        *targets,
        whitelist=None,
        blacklist=None,
        scan_id=None,
        name=None,
        modules=None,
        output_modules=None,
        config=None,
        dispatcher=None,
        strict_scope=False,
        force_start=False,
    ):
        if modules is None:
            modules = []
        if output_modules is None:
            output_modules = ["human"]
        if config is None:
            config = OmegaConf.create({})
        self.config = config
        if name is None:
            self.name = random_name()
        else:
            self.name = str(name)
        self.strict_scope = strict_scope
        self.force_start = force_start

        if scan_id is not None:
            self.id = str(scan_id)
        else:
            self.id = f"SCAN:{sha1(rand_string(20)).hexdigest()}"
        self._status = "NOT_STARTED"
        self._status_code = 0

        # Set up thread pools
        max_workers = max(1, self.config.get("max_threads", 100))
        # Shared thread pool, for module use
        self._thread_pool = ThreadPoolWrapper(concurrent.futures.ThreadPoolExecutor(max_workers=max_workers))
        # Event thread pool, for event construction, initialization
        self._event_thread_pool = ThreadPoolWrapper(concurrent.futures.ThreadPoolExecutor(max_workers=max_workers * 2))
        # Internal thread pool, for handle_event(), module setup, cleanup callbacks, etc.
        self._internal_thread_pool = ThreadPoolWrapper(concurrent.futures.ThreadPoolExecutor(max_workers=max_workers))
        self.process_pool = ThreadPoolWrapper(concurrent.futures.ProcessPoolExecutor())

        self.helpers = ConfigAwareHelper(config=self.config, scan=self)
        output_dir = self.config.get("output_dir", "")
        if output_dir:
            self.home = Path(output_dir).resolve() / self.name
        else:
            self.home = self.helpers.bbot_home / "scans" / self.name

        self.target = ScanTarget(self, *targets, strict_scope=strict_scope)

        self.modules = OrderedDict({})
        self._scan_modules = modules
        self._internal_modules = list(self._internal_modules())
        self._output_modules = output_modules
        self._modules_loaded = False

        if not whitelist:
            self.whitelist = self.target.copy()
        else:
            self.whitelist = ScanTarget(self, *whitelist, strict_scope=strict_scope)
        if not blacklist:
            blacklist = []
        self.blacklist = ScanTarget(self, *blacklist)

        if dispatcher is None:
            self.dispatcher = Dispatcher()
        else:
            self.dispatcher = dispatcher
        self.dispatcher.set_scan(self)

        self.manager = ScanManager(self)
        self.stats = ScanStats(self)

        # prevent too many brute force modules from running at one time
        # because they can bypass the global thread limit
        self.max_brute_forcers = int(self.config.get("max_brute_forcers", 1))
        self._brute_lock = threading.Semaphore(self.max_brute_forcers)

        # scope distance
        self.scope_search_distance = max(0, int(self.config.get("scope_search_distance", 1)))
        self.dns_search_distance = max(
            self.scope_search_distance, int(self.config.get("scope_dns_search_distance", 3))
        )
        self.scope_report_distance = int(self.config.get("scope_report_distance", 1))

        self._prepped = False
        self._cleanedup = False

    def prep(self):
        self.helpers.mkdir(self.home)
        if not self._prepped:
            start_msg = f"Scan with {len(self._scan_modules):,} modules seeded with {len(self.target)} targets"
            details = []
            if self.whitelist != self.target:
                details.append(f"{len(self.whitelist):,} in whitelist")
            if self.blacklist:
                details.append(f"{len(self.blacklist):,} in blacklist")
            if details:
                start_msg += f" ({', '.join(details)})"
            self.hugeinfo(start_msg)

            self.load_modules()

            self.info(f"Setting up modules...")
            self.setup_modules()

            self.success(f"Setup succeeded for {len(self.modules):,} modules.")
            self._prepped = True

    def start(self):

        self.prep()

        failed = True

        if not self.target:
            self.warning(f"No scan targets specified")

        try:
            self.status = "STARTING"

            if not self.modules:
                self.error(f"No modules loaded")
                self.status = "FAILED"
                return
            else:
                self.hugesuccess(f"Starting scan {self.name}")

            if self.stopping:
                return

            # distribute seed events
            self.manager.init_events()

            if self.stopping:
                return

            self.status = "RUNNING"
            self.start_modules()
            self.verbose(f"{len(self.modules):,} modules started")

            if self.stopping:
                return

            self.manager.loop_until_finished()
            failed = False

        except KeyboardInterrupt:
            self.stop()
            failed = False

        except ScanCancelledError:
            self.debug("Scan cancelled")

        except ScanError as e:
            self.error(f"{e}")

        except BBOTError as e:
            import traceback

            self.critical(f"Error during scan: {e}")
            self.debug(traceback.format_exc())

        except Exception:
            import traceback

            self.critical(f"Unexpected error during scan:\n{traceback.format_exc()}")

        finally:

            self.cleanup()
            self.shutdown_threadpools(wait=True)

            log_fn = self.hugesuccess
            if self.status == "ABORTING":
                self.status = "ABORTED"
                log_fn = self.hugewarning
            elif failed:
                self.status = "FAILED"
                log_fn = self.critical
            else:
                self.status = "FINISHED"

            log_fn(f"Scan {self.name} completed with status {self.status}")

            self.dispatcher.on_finish(self)

    def start_modules(self):
        self.verbose(f"Starting module threads")
        for module_name, module in self.modules.items():
            module.start()

    def setup_modules(self, remove_failed=True):
        self.load_modules()
        self.verbose(f"Setting up modules")
        hard_failed = []
        soft_failed = []
        setup_futures = dict()

        for module_name, module in self.modules.items():
            future = self._internal_thread_pool.submit_task(module._setup)
            setup_futures[future] = module_name
        for future in self.helpers.as_completed(setup_futures):
            module_name = setup_futures[future]
            status, msg = future.result()
            if status == True:
                self.debug(f"Setup succeeded for {module_name} ({msg})")
            elif status == False:
                self.error(f"Setup hard-failed for {module_name}: {msg}")
                self.modules[module_name].set_error_state()
                hard_failed.append(module_name)
            else:
                self.warning(f"Setup soft-failed for {module_name}: {msg}")
                soft_failed.append(module_name)
            if not status and remove_failed:
                self.modules.pop(module_name)

        num_output_modules = len([m for m in self.modules.values() if m._type == "output"])
        if num_output_modules < 1:
            raise ScanError("Failed to load output modules. Aborting.")
        total_failed = len(hard_failed + soft_failed)
        if hard_failed:
            msg = f"Setup hard-failed for {len(hard_failed):,} modules ({','.join(hard_failed)})"
            self.fail_setup(msg)
        elif total_failed > 0:
            self.warning(f"Setup failed for {total_failed:,} modules")

    def stop(self, wait=False):
        if self.status != "ABORTING":
            self.status = "ABORTING"
            self.hugewarning(f"Aborting scan")
            for i in range(max(10, self.max_brute_forcers * 10)):
                self._brute_lock.release()
            self.helpers.kill_children()
            self.shutdown_threadpools(wait=False)
            self.helpers.kill_children()

    def shutdown_threadpools(self, wait=True):
        pools = [
            self.process_pool,
            self._internal_thread_pool,
            self.helpers.dns._thread_pool,
            self._event_thread_pool,
            self._thread_pool,
        ]
        self.debug(f"Shutting down thread pools with wait={wait}")
        threads = []
        for pool in pools:
            t = threading.Thread(target=pool.shutdown, kwargs={"wait": False, "cancel_futures": True}, daemon=True)
            t.start()
            threads.append(t)
        if wait:
            for t in threads:
                t.join()
        if wait:
            for pool in pools:
                pool.shutdown(wait=True)
        self.debug("Finished shutting down thread pools")

    def cleanup(self):
        # clean up modules
        self.status = "CLEANING_UP"
        for mod in self.modules.values():
            mod._cleanup()
        if not self._cleanedup:
            self._cleanedup = True
            with suppress(Exception):
                self.home.rmdir()
            self.helpers.clean_old_scans()

    def in_scope(self, e):
        """
        Checks whitelist and blacklist, also taking scope_distance into account
        """
        try:
            e = make_event(e, dummy=True)
        except ValidationError:
            return False
        in_scope = e.scope_distance == 0 or self.whitelisted(e)
        return in_scope and not self.blacklisted(e)

    def blacklisted(self, e):
        e = make_event(e, dummy=True)
        return e in self.blacklist

    def whitelisted(self, e):
        e = make_event(e, dummy=True)
        return e in self.whitelist

    @property
    def word_cloud(self):
        return self.helpers.word_cloud

    @property
    def stopping(self):
        return not self.running

    @property
    def running(self):
        return 0 < self._status_code < 4

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        """
        Block setting after status has been aborted
        """
        status = str(status).strip().upper()
        if status in self._status_codes:
            if self.status == "ABORTING" and not status == "ABORTED":
                self.debug(f'Attempt to set invalid status "{status}" on aborted scan')
            else:
                self._status = status
                self._status_code = self._status_codes[status]
                self.dispatcher.on_status(self._status, self.id)
        else:
            self.debug(f'Attempt to set invalid status "{status}" on scan')

    @property
    def status_detailed(self):
        main_tasks = self._thread_pool.num_tasks
        dns_tasks = self.helpers.dns._thread_pool.num_tasks
        event_threadpool_tasks = self._event_thread_pool.num_tasks
        event_tasks = self.manager.event_queue.qsize()
        internal_tasks = self._internal_thread_pool.num_tasks
        process_tasks = self.process_pool.num_tasks
        total_tasks = main_tasks + dns_tasks + event_tasks + internal_tasks
        status = {
            "queued_tasks": {
                "main": main_tasks,
                "dns": dns_tasks,
                "internal": internal_tasks,
                "process": process_tasks,
                "event": event_threadpool_tasks,
                "total": total_tasks,
            },
            "queued_events": {
                "manager": event_tasks,
            },
        }
        return status

    def make_event(self, *args, **kwargs):
        kwargs["scan"] = self
        event = make_event(*args, **kwargs)
        return event

    @property
    def log(self):
        if self._log is None:
            self._log = logging.getLogger(f"bbot.agent.scanner")
        return self._log

    @property
    def root_event(self):
        root_event = self.make_event(data=f"{self.name} ({self.id})", event_type="SCAN", dummy=True)
        root_event._id = self.id
        root_event.scope_distance = 0
        root_event._resolved.set()
        root_event.source = root_event
        root_event.module = self.helpers._make_dummy_module(name="TARGET", _type="TARGET")
        return root_event

    @property
    def useragent(self):
        return self.config.get("user_agent", "BBOT")

    @property
    def json(self):
        j = dict()
        for i in ("id", "name"):
            v = getattr(self, i, "")
            if v:
                j.update({i: v})
        if self.target:
            j.update({"targets": [str(e.data) for e in self.target]})
        if self.whitelist:
            j.update({"whitelist": [str(e.data) for e in self.whitelist]})
        if self.blacklist:
            j.update({"blacklist": [str(e.data) for e in self.blacklist]})
        if self.modules:
            j.update({"modules": [str(m) for m in self.modules]})
        return j

    def debug(self, *args, **kwargs):
        log.debug(*args, extra={"scan_id": self.id}, **kwargs)

    def verbose(self, *args, **kwargs):
        log.verbose(*args, extra={"scan_id": self.id}, **kwargs)

    def hugeverbose(self, *args, **kwargs):
        log.hugeverbose(*args, extra={"scan_id": self.id}, **kwargs)

    def info(self, *args, **kwargs):
        log.info(*args, extra={"scan_id": self.id}, **kwargs)

    def hugeinfo(self, *args, **kwargs):
        log.hugeinfo(*args, extra={"scan_id": self.id}, **kwargs)

    def success(self, *args, **kwargs):
        log.success(*args, extra={"scan_id": self.id}, **kwargs)

    def hugesuccess(self, *args, **kwargs):
        log.hugesuccess(*args, extra={"scan_id": self.id}, **kwargs)

    def warning(self, *args, **kwargs):
        log.warning(*args, extra={"scan_id": self.id}, **kwargs)

    def hugewarning(self, *args, **kwargs):
        log.hugewarning(*args, extra={"scan_id": self.id}, **kwargs)

    def error(self, *args, **kwargs):
        log.error(*args, extra={"scan_id": self.id}, **kwargs)

    def critical(self, *args, **kwargs):
        log.critical(*args, extra={"scan_id": self.id}, **kwargs)

    def _internal_modules(self):
        for modname in module_loader.preloaded(type="internal"):
            if self.config.get(modname, True):
                yield modname

    def load_modules(self):

        if not self._modules_loaded:

            all_modules = list(set(self._scan_modules + self._output_modules + self._internal_modules))
            if not all_modules:
                self.warning(f"No modules to load")
                return

            if not self._scan_modules:
                self.warning(f"No scan modules to load")

            # install module dependencies
            succeeded, failed = self.helpers.depsinstaller.install(
                *self._scan_modules, *self._output_modules, *self._internal_modules
            )
            if failed:
                msg = f"Failed to install dependencies for {len(failed):,} modules: {','.join(failed)}"
                self.fail_setup(msg)
            modules = [m for m in self._scan_modules if m in succeeded]
            output_modules = [m for m in self._output_modules if m in succeeded]
            internal_modules = [m for m in self._internal_modules if m in succeeded]

            # Load scan modules
            self.verbose(f"Loading {len(modules):,} scan modules: {','.join(list(modules))}")
            loaded_modules, failed = self._load_modules(modules)
            self.modules.update(loaded_modules)
            if len(failed) > 0:
                msg = f"Failed to load {len(failed):,} scan modules: {','.join(failed)}"
                self.fail_setup(msg)
            if loaded_modules:
                self.info(
                    f"Loaded {len(loaded_modules):,}/{len(self._scan_modules):,} scan modules ({','.join(list(loaded_modules))})"
                )

            # Load internal modules
            self.verbose(f"Loading {len(internal_modules):,} internal modules: {','.join(list(internal_modules))}")
            loaded_internal_modules, failed_internal = self._load_modules(internal_modules)
            self.modules.update(loaded_internal_modules)
            if len(failed_internal) > 0:
                msg = f"Failed to load {len(loaded_internal_modules):,} internal modules: {','.join(loaded_internal_modules)}"
                self.fail_setup(msg)
            if loaded_internal_modules:
                self.info(
                    f"Loaded {len(loaded_internal_modules):,}/{len(self._internal_modules):,} internal modules ({','.join(list(loaded_internal_modules))})"
                )

            # Load output modules
            self.verbose(f"Loading {len(output_modules):,} output modules: {','.join(list(output_modules))}")
            loaded_output_modules, failed_output = self._load_modules(output_modules)
            self.modules.update(loaded_output_modules)
            if len(failed_output) > 0:
                msg = f"Failed to load {len(failed_output):,} output modules: {','.join(failed_output)}"
                self.fail_setup(msg)
            if loaded_output_modules:
                self.info(
                    f"Loaded {len(loaded_output_modules):,}/{len(self._output_modules):,} output modules, ({','.join(list(loaded_output_modules))})"
                )

            self.modules = OrderedDict(sorted(self.modules.items(), key=lambda x: getattr(x[-1], "_priority", 0)))
            self._modules_loaded = True

    def fail_setup(self, msg):
        msg = str(msg)
        if not self.force_start:
            msg += " (--force to run module anyway)"
        if self.force_start:
            self.error(msg)
        else:
            raise ScanError(msg)

    def _load_modules(self, modules):

        modules = [str(m) for m in modules]
        loaded_modules = {}
        failed = set()
        for module_name, module_class in module_loader.load_modules(modules).items():
            if module_class:
                try:
                    loaded_modules[module_name] = module_class(self)
                    self.verbose(f'Loaded module "{module_name}"')
                    continue
                except Exception:
                    import traceback

                    self.warning(f"Failed to load module {module_class}")
                    self.debug(traceback.format_exc())
            else:
                self.warning(f'Failed to load unknown module "{module_name}"')
            failed.add(module_name)
        return loaded_modules, failed



import os
import sys
import json
import logging
import ipaddress
from time import sleep

import bbot.core.logger  # noqa: F401
from bbot.core.errors import *
from .bbot_fixtures import *  # noqa: F401
from bbot.modules import module_loader

log = logging.getLogger(f"bbot.test")

# silence stdout
root_logger = logging.getLogger()
for h in root_logger.handlers:
    h.addFilter(lambda x: x.levelno != 100)

os.environ["BBOT_SUDO_PASS"] = "nah"

available_modules = list(module_loader.configs(type="scan"))
available_output_modules = list(module_loader.configs(type="output"))


def test_events(events, scan, helpers, bbot_config):

    assert events.ipv4.type == "IP_ADDRESS"
    assert events.ipv6.type == "IP_ADDRESS"
    assert events.netv4.type == "IP_RANGE"
    assert events.netv6.type == "IP_RANGE"
    assert events.domain.type == "DNS_NAME"
    assert "domain" in events.domain.tags
    assert events.subdomain.type == "DNS_NAME"
    assert "subdomain" in events.subdomain.tags
    assert events.open_port.type == "OPEN_TCP_PORT"
    assert events.url_unverified.type == "URL_UNVERIFIED"
    assert events.ipv4_url_unverified.type == "URL_UNVERIFIED"
    assert events.ipv6_url_unverified.type == "URL_UNVERIFIED"
    assert "" not in events.ipv4
    assert None not in events.ipv4
    assert 1 not in events.ipv4
    assert False not in events.ipv4

    # ip tests
    assert events.ipv4 == scan.make_event("8.8.8.8", dummy=True)
    assert "8.8.8.8" in events.ipv4
    assert "8.8.8.8" == events.ipv4
    assert "8.8.8.8" in events.netv4
    assert "8.8.8.9" not in events.ipv4
    assert "8.8.9.8" not in events.netv4
    assert "8.8.8.8/31" in events.netv4
    assert "8.8.8.8/30" in events.netv4
    assert "8.8.8.8/29" not in events.netv4
    assert "2001:4860:4860::8888" in events.ipv6
    assert "2001:4860:4860::8888" in events.netv6
    assert "2001:4860:4860::8889" not in events.ipv6
    assert "2002:4860:4860::8888" not in events.netv6
    assert "2001:4860:4860::8888/127" in events.netv6
    assert "2001:4860:4860::8888/126" in events.netv6
    assert "2001:4860:4860::8888/125" not in events.netv6
    assert events.emoji not in events.ipv4
    assert events.emoji not in events.netv6
    assert events.netv6 not in events.emoji
    assert "dead::c0de" == scan.make_event(" [DEaD::c0De]:88", "DNS_NAME", dummy=True)

    # hostname tests
    assert events.domain.host == "publicapis.org"
    assert events.subdomain.host == "api.publicapis.org"
    assert events.domain.host_stem == "publicapis"
    assert events.subdomain.host_stem == "api.publicapis"
    assert "api.publicapis.org" in events.domain
    assert "api.publicapis.org" in events.subdomain
    assert "fsocie.ty" not in events.domain
    assert "fsocie.ty" not in events.subdomain
    assert events.subdomain in events.domain
    assert events.domain not in events.subdomain
    assert not events.ipv4 in events.domain
    assert not events.netv6 in events.domain
    assert events.emoji not in events.domain
    assert events.domain not in events.emoji
    assert "evilcorp.com" == scan.make_event(" eViLcorp.COM.:88", "DNS_NAME", dummy=True)

    # url tests
    assert scan.make_event("http://evilcorp.com", dummy=True) == scan.make_event("http://evilcorp.com/", dummy=True)
    assert events.url_unverified.host == "api.publicapis.org"
    assert events.url_unverified in events.domain
    assert events.url_unverified in events.subdomain
    assert "api.publicapis.org:443" in events.url_unverified
    assert "publicapis.org" not in events.url_unverified
    assert events.ipv4_url_unverified in events.ipv4
    assert events.ipv4_url_unverified in events.netv4
    assert events.ipv6_url_unverified in events.ipv6
    assert events.ipv6_url_unverified in events.netv6
    assert events.emoji not in events.url_unverified
    assert events.emoji not in events.ipv6_url_unverified
    assert events.url_unverified not in events.emoji
    assert "https://evilcorp.com" == scan.make_event("https://evilcorp.com:443", dummy=True)
    assert "http://evilcorp.com" == scan.make_event("http://evilcorp.com:80", dummy=True)
    assert "http://evilcorp.com:80/asdf.js" in scan.make_event("http://evilcorp.com/asdf.js", dummy=True)
    assert "http://evilcorp.com/asdf.js" in scan.make_event("http://evilcorp.com:80/asdf.js", dummy=True)
    assert "https://evilcorp.com:443" == scan.make_event("https://evilcorp.com", dummy=True)
    assert "http://evilcorp.com:80" == scan.make_event("http://evilcorp.com", dummy=True)
    assert "https://evilcorp.com:80" == scan.make_event("https://evilcorp.com:80", dummy=True)
    assert "http://evilcorp.com:443" == scan.make_event("http://evilcorp.com:443", dummy=True)
    assert scan.make_event("https://evilcorp.com", dummy=True).with_port().geturl() == "https://evilcorp.com:443/"
    assert scan.make_event("https://evilcorp.com:666", dummy=True).with_port().geturl() == "https://evilcorp.com:666/"
    assert scan.make_event("https://[bad::c0de]", dummy=True).with_port().geturl() == "https://[bad::c0de]:443/"
    assert scan.make_event("https://[bad::c0de]:666", dummy=True).with_port().geturl() == "https://[bad::c0de]:666/"
    assert "status-200" in scan.make_event("https://evilcorp.com", "URL", events.ipv4_url, tags=["status-200"]).tags
    with pytest.raises(ValidationError, match=".*status tag.*"):
        scan.make_event("https://evilcorp.com", "URL", events.ipv4_url)

    # http response
    assert events.http_response.host == "example.com"
    assert events.http_response.port == 80
    assert events.http_response.parsed.scheme == "http"
    assert events.http_response.with_port().geturl() == "http://example.com:80/"

    # open port tests
    assert events.open_port in events.domain
    assert "api.publicapis.org:443" in events.open_port
    assert "bad.publicapis.org:443" not in events.open_port
    assert "publicapis.org:443" not in events.open_port
    assert events.ipv4_open_port in events.ipv4
    assert events.ipv4_open_port in events.netv4
    assert "8.8.8.9" not in events.ipv4_open_port
    assert events.ipv6_open_port in events.ipv6
    assert events.ipv6_open_port in events.netv6
    assert "2002:4860:4860::8888" not in events.ipv6_open_port
    assert events.emoji not in events.ipv6_open_port
    assert events.ipv6_open_port not in events.emoji

    # attribute tests
    assert events.ipv4.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4.port is None
    assert events.ipv6.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6.port is None
    assert events.domain.port is None
    assert events.subdomain.port is None
    assert events.open_port.host == "api.publicapis.org"
    assert events.open_port.port == 443
    assert events.ipv4_open_port.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4_open_port.port == 443
    assert events.ipv6_open_port.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6_open_port.port == 443
    assert events.url_unverified.host == "api.publicapis.org"
    assert events.url_unverified.port == 443
    assert events.ipv4_url_unverified.host == ipaddress.ip_address("8.8.8.8")
    assert events.ipv4_url_unverified.port == 443
    assert events.ipv6_url_unverified.host == ipaddress.ip_address("2001:4860:4860::8888")
    assert events.ipv6_url_unverified.port == 443

    javascript_event = scan.make_event("http://evilcorp.com/asdf/a.js?b=c#d", "URL_UNVERIFIED", dummy=True)
    assert "extension-js" in javascript_event.tags
    assert "httpx-only" in javascript_event.tags

    # scope distance
    event1 = scan.make_event("1.2.3.4", dummy=True)
    assert event1._scope_distance == -1
    event1.make_in_scope()
    assert event1._scope_distance == 0
    event2 = scan.make_event("2.3.4.5", source=event1)
    assert event2._scope_distance == 1
    event3 = scan.make_event("3.4.5.6", source=event2)
    assert event3._scope_distance == 2

    # internal event tracking
    root_event = scan.make_event("0.0.0.0", dummy=True)
    internal_event1 = scan.make_event("1.2.3.4", source=root_event, internal=True)
    assert internal_event1._internal == True
    assert internal_event1._made_internal == True
    internal_event1.make_in_scope()
    assert internal_event1._internal == False
    assert internal_event1._made_internal == False
    internal_event2 = scan.make_event("2.3.4.5", source=internal_event1, internal=True)
    internal_event3 = scan.make_event("3.4.5.6", source=internal_event2, internal=True)
    internal_event4 = scan.make_event("4.5.6.7", source=internal_event3)
    source_trail = internal_event4.make_in_scope()
    assert internal_event4._internal == False
    assert internal_event3._internal == False
    assert internal_event2._internal == False
    assert len(source_trail) == 2
    assert internal_event2 in source_trail
    assert internal_event3 in source_trail

    # event sorting
    sort1 = scan.make_event("127.0.0.1", dummy=True)
    sort1._priority = 1
    sort2 = scan.make_event("127.0.0.1", dummy=True)
    sort2._priority = 2
    sort3 = scan.make_event("127.0.0.1", dummy=True)
    sort3._priority = 3
    mod1 = helpers._make_dummy_module(name="MOD1", _type="ASDF")
    mod1._priority = 1
    mod2 = helpers._make_dummy_module(name="MOD2", _type="ASDF")
    mod2._priority = 2
    mod3 = helpers._make_dummy_module(name="MOD3", _type="ASDF")
    mod3._priority = 3
    sort1.module = mod1
    sort2.module = mod2
    sort3.module = mod3
    assert 2 < sort1.priority < 2.01
    assert sort1 < sort2
    assert sort1 < sort3
    assert 4 < sort2.priority < 4.01
    assert sort2 > sort1
    assert sort2 < sort3
    assert 6 < sort3.priority < 6.01
    assert sort3 > sort1
    assert sort3 > sort2
    assert tuple(sorted([sort3, sort2, sort1])) == (sort1, sort2, sort3)

    # test validation
    test_vuln = scan.make_event(
        {"host": "EVILcorp.com", "severity": "iNfo ", "description": "asdf"}, "VULNERABILITY", dummy=True
    )
    assert test_vuln.data["host"] == "evilcorp.com"
    assert test_vuln.data["severity"] == "INFO"
    test_vuln2 = scan.make_event(
        {"host": "192.168.1.1", "severity": "iNfo ", "description": "asdf"}, "VULNERABILITY", dummy=True
    )
    assert json.loads(test_vuln2.data_human)["severity"] == "INFO"
    assert test_vuln2.host.is_private
    with pytest.raises(ValidationError, match=".*severity.*\n.*field required.*"):
        test_vuln = scan.make_event({"host": "evilcorp.com", "description": "asdf"}, "VULNERABILITY", dummy=True)
    with pytest.raises(ValidationError, match=".*host.*\n.*Invalid host.*"):
        test_vuln = scan.make_event(
            {"host": "!@#$", "severity": "INFO", "description": "asdf"}, "VULNERABILITY", dummy=True
        )
    with pytest.raises(ValidationError, match=".*severity.*\n.*Invalid severity.*"):
        test_vuln = scan.make_event(
            {"host": "evilcorp.com", "severity": "WACK", "description": "asdf"}, "VULNERABILITY", dummy=True
        )


def test_manager(bbot_config):
    from bbot.scanner import Scanner

    # test _emit_event
    results = []
    success_callback = lambda e: results.append("success")
    scan1 = Scanner("127.0.0.1", config=bbot_config)
    scan1.status = "RUNNING"
    scan1.manager.queue_event = lambda e: results.append(e)
    manager = scan1.manager
    localhost = scan1.make_event("127.0.0.1", source=scan1.root_event)

    class DummyModule1:
        _type = "output"
        suppress_dupes = True

    class DummyModule2:
        _type = "DNS"
        suppress_dupes = True

    localhost.module = DummyModule1()
    # test abort_if
    manager._emit_event(localhost, abort_if=lambda e: e.module._type == "output")
    assert len(results) == 0
    manager._emit_event(
        localhost, on_success_callback=success_callback, abort_if=lambda e: e.module._type == "plumbus"
    )
    assert localhost in results
    assert "success" in results
    results.clear()
    # test deduplication
    manager._emit_event(localhost, on_success_callback=success_callback)
    assert len(results) == 0
    # test dns resolution
    googledns = scan1.make_event("8.8.8.8", source=scan1.root_event)
    googledns.module = DummyModule2()
    googledns.source = "asdf"
    googledns.make_in_scope()
    event_children = []
    manager.emit_event = lambda e, *args, **kwargs: event_children.append(e)
    manager._emit_event(googledns)
    assert len(event_children) > 0
    assert googledns in results
    results.clear()
    event_children.clear()
    # same dns event
    manager._emit_event(googledns)
    assert len(results) == 0
    assert len(event_children) == 0
    # same dns event but with _force_output
    googledns._force_output = True
    manager._emit_event(googledns)
    assert googledns in results
    assert len(event_children) == 0
    googledns._force_output = False
    results.clear()
    # same dns event but different source
    googledns.source_id = "fdsa"
    manager._emit_event(googledns)
    assert len(event_children) == 0
    assert googledns in results

    # event filtering based on scope_distance
    output_queue = []
    module_queue = []
    manager_queue = []
    scan1 = Scanner("127.0.0.1", modules=["ipneighbor"], output_modules=["json"], config=bbot_config)
    scan1.status = "RUNNING"
    scan1.load_modules()
    manager = scan1.manager
    test_event1 = scan1.make_event("1.2.3.4", source=scan1.root_event)
    test_event1.make_in_scope()
    test_event2 = scan1.make_event("2.3.4.5", source=test_event1)
    test_event3 = scan1.make_event("3.4.5.6", source=test_event2)

    scan1.manager.queue_event = lambda e: manager_queue.append(e)

    scan1.scope_search_distance = 1
    scan1.scope_report_distance = 0
    assert test_event1.scope_distance == 0
    manager._emit_event(test_event1)
    assert test_event1 in manager_queue
    assert test_event1._internal == False
    assert test_event2.scope_distance == 1
    manager._emit_event(test_event2)
    assert test_event2 in manager_queue
    assert test_event2._internal == True
    manager_queue.clear()
    manager.events_accepted.clear()

    scan1.modules["json"].queue_event = lambda e: output_queue.append(e)
    scan1.modules["ipneighbor"].queue_event = lambda e: module_queue.append(e)

    # in-scope event
    assert test_event1.scope_distance == 0
    manager.distribute_event(test_event1)
    assert hash(test_event1) in manager.events_distributed
    assert test_event1 in module_queue
    assert test_event1 in output_queue
    module_queue.clear()
    output_queue.clear()
    # duplicate event
    manager.distribute_event(test_event1)
    assert test_event1 not in module_queue
    assert test_event1 in output_queue
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 1
    assert test_event2.scope_distance == 1
    manager.distribute_event(test_event2)
    assert test_event2 in module_queue
    assert test_event2 not in output_queue
    assert test_event2._internal == True
    assert test_event2._force_output == False
    assert scan1.modules["json"]._filter_event(test_event2) == False
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 2
    assert test_event3.scope_distance == 2
    manager.distribute_event(test_event3)
    assert test_event3 not in module_queue
    assert test_event3 not in output_queue
    module_queue.clear()
    output_queue.clear()
    manager.events_distributed.clear()
    # event.scope_distance == 2 and _force_output == True
    test_event3._force_output = True
    assert test_event3.scope_distance == 2
    manager.distribute_event(test_event3)
    assert test_event3 not in module_queue
    assert test_event3 in output_queue


def test_curl(helpers):

    helpers.curl()
    helpers.curl(url="http://www.example.com", ignore_bbot_global_settings=True)
    helpers.curl(url="http://www.example.com", head_mode=True)
    helpers.curl(url="http://www.example.com", raw_body=True)
    helpers.curl(
        url="http://www.example.com",
        raw_path=True,
        headers={"test": "test", "test2": ["test2"]},
        ignore_bbot_global_settings=False,
        post_data={"test": "test"},
        method="POST",
        cookies={"test": "test"},
        path_override="/index.html",
    )


def test_helpers(patch_requests, patch_commands, helpers, scan):

    old_run, old_run_live = patch_commands
    request, download = patch_requests

    ### URL ###
    bad_urls = (
        "http://e.co/index.html",
        "http://e.co/u/1111/info",
        "http://e.co/u/2222/info",
        "http://e.co/u/3333/info",
        "http://e.co/u/4444/info",
        "http://e.co/u/5555/info",
    )
    new_urls = tuple(helpers.collapse_urls(bad_urls, threshold=4))
    assert len(new_urls) == 2
    new_urls = tuple(sorted([u.geturl() for u in helpers.collapse_urls(bad_urls, threshold=5)]))
    assert new_urls == bad_urls

    new_url = helpers.add_get_params("http://evilcorp.com/a?p=1&q=2", {"r": 3, "s": "asdf"}).geturl()
    query = dict(s.split("=") for s in new_url.split("?")[-1].split("&"))
    query = tuple(sorted(query.items(), key=lambda x: x[0]))
    assert query == (
        ("p", "1"),
        ("q", "2"),
        ("r", "3"),
        ("s", "asdf"),
    )
    assert tuple(sorted(helpers.get_get_params("http://evilcorp.com/a?p=1&q=2#frag").items())) == (
        ("p", ["1"]),
        ("q", ["2"]),
    )

    assert helpers.clean_url("http://evilcorp.com:80").geturl() == "http://evilcorp.com/"
    assert helpers.clean_url("http://evilcorp.com/asdf?a=asdf#frag").geturl() == "http://evilcorp.com/asdf"
    assert helpers.clean_url("http://evilcorp.com//asdf").geturl() == "http://evilcorp.com/asdf"

    assert helpers.url_depth("http://evilcorp.com/asdf/user/") == 2
    assert helpers.url_depth("http://evilcorp.com/asdf/user") == 2
    assert helpers.url_depth("http://evilcorp.com/asdf/") == 1
    assert helpers.url_depth("http://evilcorp.com/asdf") == 1
    assert helpers.url_depth("http://evilcorp.com/") == 0
    assert helpers.url_depth("http://evilcorp.com") == 0

    ### HTTP COMPARE ###
    compare_helper = helpers.http_compare("http://www.example.com")
    compare_helper.compare("http://www.example.com", headers={"asdf": "asdf"})
    compare_helper.compare("http://www.example.com", cookies={"asdf": "asdf"})
    compare_helper.compare("http://www.example.com", check_reflection=True)
    compare_helper.compare_body({"asdf": "fdsa"}, {"fdsa": "asdf"})
    for mode in ("getparam", "header", "cookie"):
        compare_helper.canary_check("http://www.example.com", mode=mode) == True

    ### MISC ###
    assert helpers.is_domain("evilcorp.co.uk")
    assert not helpers.is_domain("www.evilcorp.co.uk")
    assert helpers.is_subdomain("www.evilcorp.co.uk")
    assert not helpers.is_subdomain("evilcorp.co.uk")
    assert helpers.is_url("http://evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert helpers.is_url("https://evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert not helpers.is_url("https:/evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert not helpers.is_url("/evilcorp.co.uk/asdf?a=b&c=d#asdf")
    assert not helpers.is_url("ftp://evilcorp.co.uk")
    assert helpers.parent_domain("www.evilcorp.co.uk") == "evilcorp.co.uk"
    assert helpers.parent_domain("evilcorp.co.uk") == "evilcorp.co.uk"
    assert helpers.parent_domain("localhost") == "localhost"
    assert list(helpers.domain_parents("test.www.evilcorp.co.uk")) == ["www.evilcorp.co.uk", "evilcorp.co.uk"]
    assert list(helpers.domain_parents("www.evilcorp.co.uk", include_self=True)) == [
        "www.evilcorp.co.uk",
        "evilcorp.co.uk",
    ]
    assert list(helpers.domain_parents("evilcorp.co.uk", include_self=True)) == ["evilcorp.co.uk"]
    assert list(helpers.ip_network_parents("0.0.0.0/2")) == [
        ipaddress.ip_network("0.0.0.0/1"),
        ipaddress.ip_network("0.0.0.0/0"),
    ]
    assert list(helpers.ip_network_parents("0.0.0.0/1", include_self=True)) == [
        ipaddress.ip_network("0.0.0.0/1"),
        ipaddress.ip_network("0.0.0.0/0"),
    ]
    assert helpers.is_ip("127.0.0.1")
    assert not helpers.is_ip("127.0.0.0.1")

    assert helpers.domain_stem("evilcorp.co.uk") == "evilcorp"
    assert helpers.domain_stem("www.evilcorp.co.uk") == "www.evilcorp"

    assert helpers.host_in_host("www.evilcorp.com", "evilcorp.com") == True
    assert helpers.host_in_host("asdf.www.evilcorp.com", "evilcorp.com") == True
    assert helpers.host_in_host("evilcorp.com", "www.evilcorp.com") == False
    assert helpers.host_in_host("evilcorp.com", "evilcorp.com") == True
    assert helpers.host_in_host("evilcorp.com", "eevilcorp.com") == False
    assert helpers.host_in_host("eevilcorp.com", "evilcorp.com") == False
    assert helpers.host_in_host("evilcorp.com", "evilcorp") == False
    assert helpers.host_in_host("evilcorp", "evilcorp.com") == False
    assert helpers.host_in_host("evilcorp.com", "com") == True

    assert tuple(helpers.extract_emails("asdf@asdf.com\nT@t.Com&a=a@a.com__ b@b.com")) == (
        "asdf@asdf.com",
        "t@t.com",
        "a@a.com",
        "b@b.com",
    )

    assert helpers.split_host_port("https://evilcorp.co.uk") == ("evilcorp.co.uk", 443)
    assert helpers.split_host_port("http://evilcorp.co.uk:666") == ("evilcorp.co.uk", 666)
    assert helpers.split_host_port("evilcorp.co.uk:666") == ("evilcorp.co.uk", 666)
    assert helpers.split_host_port("evilcorp.co.uk") == ("evilcorp.co.uk", None)
    assert helpers.split_host_port("d://wat:wat") == ("wat", None)
    assert helpers.split_host_port("https://[dead::beef]:8338") == (ipaddress.ip_address("dead::beef"), 8338)
    extracted_words = helpers.extract_words("blacklanternsecurity")
    assert "black" in extracted_words
    # assert "blacklantern" in extracted_words
    # assert "lanternsecurity" in extracted_words
    # assert "blacklanternsecurity" in extracted_words
    assert "bls" in extracted_words
    ipv4_netloc = helpers.make_netloc("192.168.1.1", 80)
    assert ipv4_netloc == "192.168.1.1:80"
    ipv6_netloc = helpers.make_netloc("dead::beef", "443")
    assert ipv6_netloc == "[dead::beef]:443"

    assert helpers.get_file_extension("https://evilcorp.com/evilcorp.com/test/asdf.TXT") == "txt"
    assert helpers.get_file_extension("/etc/conf/test.tar.gz") == "gz"
    assert helpers.get_file_extension("/etc/passwd") == ""

    assert list(helpers.search_dict_by_key("asdf", {"asdf": "fdsa", 4: [{"asdf": 5}]})) == ["fdsa", 5]
    assert list(helpers.search_dict_by_key("asdf", {"wat": {"asdf": "fdsa"}})) == ["fdsa"]
    assert list(helpers.search_dict_by_key("asdf", [{"wat": {"nope": 1}}, {"wat": [{"asdf": "fdsa"}]}])) == ["fdsa"]
    assert not list(helpers.search_dict_by_key("asdf", [{"wat": {"nope": 1}}, {"wat": [{"fdsa": "asdf"}]}]))
    assert not list(helpers.search_dict_by_key("asdf", "asdf"))

    filtered_dict = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "api_key"
    )
    assert "api_key" in filtered_dict["modules"]["c99"]
    assert "filterme" not in filtered_dict["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict["modules"]

    filtered_dict2 = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "c99"
    )
    assert "api_key" in filtered_dict2["modules"]["c99"]
    assert "filterme" in filtered_dict2["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict2["modules"]

    filtered_dict3 = helpers.filter_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}},
        "key",
        fuzzy=True,
    )
    assert "api_key" in filtered_dict3["modules"]["c99"]
    assert "filterme" not in filtered_dict3["modules"]["c99"]
    assert "ipneighbor" not in filtered_dict3["modules"]

    cleaned_dict = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "api_key"
    )
    assert "api_key" not in cleaned_dict["modules"]["c99"]
    assert "filterme" in cleaned_dict["modules"]["c99"]
    assert "ipneighbor" in cleaned_dict["modules"]

    cleaned_dict2 = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}}, "c99"
    )
    assert "c99" not in cleaned_dict2["modules"]
    assert "ipneighbor" in cleaned_dict2["modules"]

    cleaned_dict3 = helpers.clean_dict(
        {"modules": {"c99": {"api_key": "1234", "filterme": "asdf"}, "ipneighbor": {"test": "test"}}},
        "key",
        fuzzy=True,
    )
    assert "api_key" not in cleaned_dict3["modules"]["c99"]
    assert "filterme" in cleaned_dict3["modules"]["c99"]
    assert "ipneighbor" in cleaned_dict3["modules"]

    replaced = helpers.search_format_dict({"asdf": [{"wat": {"here": "{replaceme}!"}}, {500: True}]}, replaceme="asdf")
    assert replaced["asdf"][1][500] == True
    assert replaced["asdf"][0]["wat"]["here"] == "asdf!"

    assert helpers.split_list([1, 2, 3, 4, 5]) == [[1, 2], [3, 4, 5]]
    assert list(helpers.grouper("ABCDEFG", 3)) == [["A", "B", "C"], ["D", "E", "F"], ["G"]]

    assert len(helpers.rand_string(3)) == 3
    assert len(helpers.rand_string(1)) == 1
    assert len(helpers.rand_string(0)) == 0
    assert type(helpers.rand_string(0)) == str

    test_file = Path(scan.config["home"]) / "testfile.asdf"
    test_file.touch()

    assert test_file.is_file()
    backup = helpers.backup_file(test_file)
    assert backup.name == "testfile.1.asdf"
    assert not test_file.exists()
    assert backup.is_file()
    test_file.touch()
    backup2 = helpers.backup_file(test_file)
    assert backup2.name == "testfile.1.asdf"
    assert not test_file.exists()
    assert backup2.is_file()
    older_backup = Path(scan.config["home"]) / "testfile.2.asdf"
    assert older_backup.is_file()
    older_backup.unlink()
    backup.unlink()

    with open(test_file, "w") as f:
        f.write("asdf\nfdsa")

    assert "asdf" in helpers.str_or_file(str(test_file))
    assert "nope" in helpers.str_or_file("nope")
    assert tuple(helpers.chain_lists([str(test_file), "nope"], try_files=True)) == ("asdf", "fdsa", "nope")
    assert test_file.is_file()

    with pytest.raises(DirectoryCreationError, match="Failed to create.*"):
        helpers.mkdir(test_file)

    helpers._rm_at_exit(test_file)
    assert not test_file.exists()

    ### VALIDATORS ###
    # hosts
    assert helpers.validators.validate_host(" evilCorp.COM") == "evilcorp.com"
    assert helpers.validators.validate_host("LOCALHOST ") == "localhost"
    assert helpers.validators.validate_host(" 192.168.1.1") == "192.168.1.1"
    assert helpers.validators.validate_host(" Dead::c0dE ") == "dead::c0de"
    assert helpers.validators.soft_validate(" evilCorp.COM", "host") == True
    assert helpers.validators.soft_validate("!@#$", "host") == False
    with pytest.raises(ValueError):
        assert helpers.validators.validate_host("!@#$")
    # ports
    assert helpers.validators.validate_port(666) == 666
    assert helpers.validators.validate_port(666666) == 65535
    assert helpers.validators.soft_validate(666, "port") == True
    assert helpers.validators.soft_validate("!@#$", "port") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_port("asdf")
    # urls
    assert helpers.validators.validate_url(" httP://evilcorP.com/asdf?a=b&c=d#e") == "http://evilcorp.com/asdf"
    assert (
        helpers.validators.validate_url_parsed(" httP://evilcorP.com/asdf?a=b&c=d#e").geturl()
        == "http://evilcorp.com/asdf"
    )
    assert helpers.validators.soft_validate(" httP://evilcorP.com/asdf?a=b&c=d#e", "url") == True
    assert helpers.validators.soft_validate("!@#$", "url") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_url("!@#$")
    # severities
    assert helpers.validators.validate_severity(" iNfo") == "INFO"
    assert helpers.validators.soft_validate(" iNfo", "severity") == True
    assert helpers.validators.soft_validate("NOPE", "severity") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_severity("NOPE")
    # emails
    assert helpers.validators.validate_email(" bOb@eViLcorp.COM") == "bob@evilcorp.com"
    assert helpers.validators.soft_validate(" bOb@eViLcorp.COM", "email") == True
    assert helpers.validators.soft_validate("!@#$", "email") == False
    with pytest.raises(ValueError):
        helpers.validators.validate_email("!@#$")

    assert type(helpers.make_date()) == str

    def raise_filenotfound():
        raise FileNotFoundError("asdf")

    def raise_brokenpipe():
        raise BrokenPipeError("asdf")

    from bbot.core.helpers import command

    command.catch(raise_filenotfound)
    command.catch(raise_brokenpipe)

    ### COMMAND ###
    assert "plumbus\n" in old_run(helpers, ["echo", "plumbus"], text=True).stdout
    assert "plumbus\n" in list(old_run_live(helpers, ["echo", "plumbus"]))
    expected_output = ["lumbus\n", "plumbus\n", "rumbus\n"]
    assert list(old_run_live(helpers, ["cat"], input="lumbus\nplumbus\nrumbus")) == expected_output

    def plumbus_generator():
        yield "lumbus"
        yield "plumbus"

    assert "plumbus\n" in list(old_run_live(helpers, ["cat"], input=plumbus_generator()))
    tempfile = helpers.tempfile(("lumbus", "plumbus"), pipe=True)
    with open(tempfile) as f:
        assert "plumbus\n" in list(f)
    tempfile = helpers.tempfile(("lumbus", "plumbus"), pipe=False)
    with open(tempfile) as f:
        assert "plumbus\n" in list(f)

    ### CACHE ###
    helpers.cache_put("string", "wat")
    helpers.cache_put("binary", b"wat")
    assert helpers.cache_get("string") == "wat"
    assert helpers.cache_get("binary") == "wat"
    assert helpers.cache_get("binary", text=False) == b"wat"
    cache_filename = helpers.cache_filename("string")
    (m, i, d, n, u, g, sz, atime, mtime, ctime) = os.stat(str(cache_filename))
    # change modified time to be 10 days in the past
    os.utime(str(cache_filename), times=(atime, mtime - (3600 * 24 * 10)))
    assert helpers.cache_get("string", cache_hrs=24 * 7) is None
    assert helpers.cache_get("string", cache_hrs=24 * 14) == "wat"

    cache_dict = helpers.CacheDict(max_size=10)
    cache_dict.put("1", 2)
    assert cache_dict["1"] == 2
    assert cache_dict.get("1") == 2
    assert len(cache_dict) == 1
    cache_dict["2"] = 3
    assert cache_dict["2"] == 3
    assert cache_dict.get("2") == 3
    assert len(cache_dict) == 2
    for i in range(20):
        cache_dict[str(i)] = i + 1
    assert len(cache_dict) == 10
    assert tuple(cache_dict) == tuple(hash(str(x)) for x in range(10, 20))

    ### WEB ###
    assert getattr(request(helpers, "https://api.publicapis.org/health"), "text", "").startswith("{")
    assert getattr(request(helpers, "https://api.publicapis.org/health", cache_for=60), "text", "").startswith("{")
    filename = download(helpers, "https://api.publicapis.org/health", cache_hrs=1)
    assert Path(str(filename)).is_file()
    assert helpers.is_cached("https://api.publicapis.org/health")

    assert helpers.wordlist("https://api.publicapis.org/healthasdf").is_file()
    test_file = Path(scan.config["home"]) / "testfile.asdf"
    with open(test_file, "w") as f:
        for i in range(100):
            f.write(f"{i}\n")
    assert len(list(open(test_file).readlines())) == 100
    assert helpers.wordlist(test_file).is_file()
    truncated_file = helpers.wordlist(test_file, lines=10)
    assert truncated_file.is_file()
    assert len(list(open(truncated_file).readlines())) == 10
    with pytest.raises(WordlistError):
        helpers.wordlist("/tmp/a9pseoysadf/asdkgjaosidf")
    test_file.unlink()

    ### DNS ###
    # resolution
    assert all([helpers.is_ip(i) for i in helpers.resolve("scanme.nmap.org")])
    assert "dns.google" in helpers.resolve("8.8.8.8")
    assert "dns.google" in helpers.resolve("2001:4860:4860::8888")
    resolved_ips = helpers.resolve("dns.google")
    assert "2001:4860:4860::8888" in resolved_ips
    assert "8.8.8.8" in resolved_ips
    assert any([helpers.is_subdomain(h) for h in helpers.resolve("google.com", type="mx")])
    v6_ips = helpers.resolve("www.google.com", type="AAAA")
    assert all([i.version == 6 for i in [ipaddress.ip_address(_) for _ in v6_ips]])
    assert not helpers.resolve(f"{helpers.rand_string(length=30)}.com")
    # batch resolution
    batch_results = list(helpers.resolve_batch(["8.8.8.8", "dns.google"]))
    assert len(batch_results) == 2
    batch_results = dict(batch_results)
    assert any([x in batch_results["dns.google"] for x in ("8.8.8.8", "8.8.4.4")])
    assert "dns.google" in batch_results["8.8.8.8"]
    # "any" type
    resolved = helpers.resolve("google.com", type="any")
    assert any([helpers.is_subdomain(h) for h in resolved])
    # wildcards
    assert helpers.is_wildcard("asdf.wat.blacklanternsecurity.github.io") == (True, "_wildcard.github.io")
    assert hash("github.io") in helpers.dns._wildcard_cache
    assert helpers.dns._wildcard_cache[hash("github.io")] == True
    assert helpers.is_wildcard("asdf.asdf.asdf.github.io") == (True, "_wildcard.github.io")
    assert helpers.is_wildcard("github.io") == (False, "github.io")
    assert helpers.is_wildcard("mail.google.com") == (False, "mail.google.com")
    wildcard_event1 = scan.make_event("wat.asdf.fdsa.github.io", "DNS_NAME", dummy=True)
    wildcard_event2 = scan.make_event("wats.asd.fdsa.github.io", "DNS_NAME", dummy=True)
    children, event_tags1, event_whitelisted1, event_blacklisted1 = scan.helpers.resolve_event(wildcard_event1)
    children, event_tags2, event_whitelisted2, event_blacklisted2 = scan.helpers.resolve_event(wildcard_event2)
    assert "wildcard" in event_tags1
    assert "wildcard" in event_tags2
    assert wildcard_event1.data == "_wildcard.github.io"
    assert wildcard_event2.data == "_wildcard.github.io"
    assert event_tags1 == event_tags2
    assert event_whitelisted1 == event_whitelisted2
    assert event_blacklisted1 == event_blacklisted2

    msg = "Ignore this error, it belongs here"

    def raise_e():
        raise Exception(msg)

    def raise_k():
        raise KeyboardInterrupt(msg)

    def raise_s():
        raise ScanCancelledError(msg)

    def raise_b():
        raise BrokenPipeError(msg)

    helpers.dns._catch_keyboardinterrupt(raise_e)
    helpers.dns._catch_keyboardinterrupt(raise_k)
    scan.manager.catch(raise_e, _on_finish_callback=raise_e)
    scan.manager.catch(raise_k)
    scan.manager.catch(raise_s)
    scan.manager.catch(raise_b)

    ## NTLM
    testheader = "TlRMTVNTUAACAAAAHgAeADgAAAAVgorilwL+bvnVipUAAAAAAAAAAJgAmABWAAAACgBjRQAAAA9XAEkATgAtAFMANAAyAE4ATwBCAEQAVgBUAEsAOAACAB4AVwBJAE4ALQBTADQAMgBOAE8AQgBEAFYAVABLADgAAQAeAFcASQBOAC0AUwA0ADIATgBPAEIARABWAFQASwA4AAQAHgBXAEkATgAtAFMANAAyAE4ATwBCAEQAVgBUAEsAOAADAB4AVwBJAE4ALQBTADQAMgBOAE8AQgBEAFYAVABLADgABwAIAHUwOZlfoNgBAAAAAA=="
    decoded = helpers.ntlm.ntlmdecode(testheader)
    assert decoded["NetBIOS_Domain_Name"] == "WIN-S42NOBDVTK8"
    assert decoded["NetBIOS_Computer_Name"] == "WIN-S42NOBDVTK8"
    assert decoded["DNS_Domain_name"] == "WIN-S42NOBDVTK8"
    assert decoded["FQDN"] == "WIN-S42NOBDVTK8"
    assert decoded["Timestamp"] == b"u09\x99_\xa0\xd8\x01"
    with pytest.raises(NTLMError):
        helpers.ntlm.ntlmdecode("asdf")

    # interact.sh
    interactsh_client = helpers.interactsh()
    with pytest.raises(InteractshError):
        interactsh_client.register()
    assert not list(interactsh_client.poll())
    with pytest.raises(InteractshError):
        interactsh_client.deregister()


def test_dns_resolvers(patch_requests, helpers):
    assert type(helpers.dns.resolvers) == set
    assert hasattr(helpers.dns.resolver_file, "is_file")
    assert hasattr(helpers.dns.mass_resolver_file, "is_file")


def test_word_cloud(helpers, bbot_config):
    number_mutations = helpers.word_cloud.get_number_mutations("base2_p013", n=5, padding=2)
    assert "base0_p013" in number_mutations
    assert "base7_p013" in number_mutations
    assert "base8_p013" not in number_mutations
    assert "base2_p008" in number_mutations
    assert "base2_p007" not in number_mutations
    assert "base2_p018" in number_mutations
    assert "base2_p0134" in number_mutations
    assert "base2_p0135" not in number_mutations

    permutations = helpers.word_cloud.mutations("_base", numbers=1)
    assert ("_base", "dev") in permutations
    assert ("dev", "_base") in permutations

    # saving and loading
    from bbot.scanner.scanner import Scanner

    scan1 = Scanner("127.0.0.1", config=bbot_config)
    word_cloud = scan1.helpers.word_cloud
    word_cloud.add_word("lantern")
    word_cloud.add_word("black")
    word_cloud.add_word("black")
    word_cloud.save()
    with open(word_cloud.default_filename) as f:
        word_cloud_content = [l.rstrip() for l in f.read().splitlines()]
    assert len(word_cloud_content) == 2
    assert "2\tblack" in word_cloud_content
    assert "1\tlantern" in word_cloud_content
    word_cloud.save(limit=1)
    with open(word_cloud.default_filename) as f:
        word_cloud_content = [l.rstrip() for l in f.read().splitlines()]
    assert len(word_cloud_content) == 1
    assert "2\tblack" in word_cloud_content
    assert "1\tlantern" not in word_cloud_content
    word_cloud.clear()
    with open(word_cloud.default_filename, "w") as f:
        f.write("plumbus\nrumbus")
    word_cloud.load()
    assert word_cloud["plumbus"] == 1
    assert word_cloud["rumbus"] == 1


def test_modules(patch_requests, patch_commands, scan, helpers, events, bbot_config):

    # base module _filter_event()
    from bbot.modules.base import BaseModule

    base_module = BaseModule(scan)
    localhost2 = scan.make_event("127.0.0.2", source=events.subdomain)
    localhost2.make_in_scope()
    # base cases
    assert base_module._filter_event("FINISHED") == True
    assert base_module._filter_event("WAT") == False
    base_module._watched_events = None
    base_module.watched_events = ["*"]
    assert base_module._filter_event("WAT") == False
    assert base_module._filter_event(events.emoji) == True
    base_module._watched_events = None
    base_module.watched_events = ["IP_ADDRESS"]
    assert base_module._filter_event(events.ipv4) == True
    assert base_module._filter_event(events.domain) == False
    assert base_module._filter_event(events.localhost) == True
    assert base_module._filter_event(localhost2) == True
    # target only
    base_module.target_only = True
    assert base_module._filter_event(localhost2) == False
    localhost2.tags.add("target")
    assert base_module._filter_event(localhost2) == True
    base_module.target_only = False
    # in scope only
    localhost3 = scan.make_event("127.0.0.2", source=events.subdomain)
    base_module.in_scope_only = True
    assert base_module._filter_event(events.localhost) == True
    assert base_module._filter_event(localhost3) == False
    base_module.in_scope_only = False
    # scope distance
    base_module.scope_distance_modifier = 0
    localhost2._scope_distance = 0
    assert base_module._filter_event(localhost2) == True
    localhost2._scope_distance = 1
    assert base_module._filter_event(localhost2) == True
    localhost2._scope_distance = 2
    assert base_module._filter_event(localhost2) == False
    localhost2._scope_distance = -1
    assert base_module._filter_event(localhost2) == False
    base_module.scope_distance_modifier = -1
    # special case for IPs and ranges
    base_module.watched_events = ["IP_ADDRESS", "IP_RANGE"]
    ip_range = scan.make_event("127.0.0.0/24", dummy=True)
    localhost4 = scan.make_event("127.0.0.1", source=ip_range)
    localhost4.make_in_scope()
    localhost4.module = "plumbus"
    assert base_module._filter_event(localhost4) == True
    localhost4.module = "speculate"
    assert base_module._filter_event(localhost4) == False

    from bbot.scanner.scanner import Scanner

    scan2 = Scanner(modules=list(available_modules), output_modules=list(available_output_modules), config=bbot_config)
    scan2.load_modules()
    scan2.status = "RUNNING"

    # attributes, descriptions, etc.
    for module_name, module in scan2.modules.items():
        # flags
        assert module._type in ("internal", "output", "scan")

    # module preloading
    all_preloaded = module_loader.preloaded()
    assert "massdns" in all_preloaded
    assert "DNS_NAME" in all_preloaded["massdns"]["watched_events"]
    assert "DNS_NAME" in all_preloaded["massdns"]["produced_events"]
    assert "subdomain-enum" in all_preloaded["massdns"]["flags"]
    assert "wordlist" in all_preloaded["massdns"]["config"]
    assert type(all_preloaded["massdns"]["config"]["max_resolvers"]) == int
    assert all_preloaded["sslcert"]["deps"]["pip"]
    assert all_preloaded["sslcert"]["deps"]["apt"]
    assert all_preloaded["massdns"]["deps"]["ansible"]

    for module_name, preloaded in all_preloaded.items():
        # either active or passive and never both
        flags = preloaded.get("flags", [])
        if preloaded["type"] == "scan":
            assert ("active" in flags and not "passive" in flags) or (
                not "active" in flags and "passive" in flags
            ), f'module "{module_name}" must have either "active" or "passive" flag'
            assert preloaded["meta"]["description"], f"{module_name} must have a description"

        # attribute checks
        watched_events = preloaded.get("watched_events")
        produced_events = preloaded.get("produced_events")

        assert type(watched_events) == list
        assert type(produced_events) == list
        assert watched_events, f"{module_name}.watched_events must not be empty"
        assert type(watched_events) == list, f"{module_name}.watched_events must be of type list"
        assert type(produced_events) == list, f"{module_name}.produced_events must be of type list"
        assert all(
            [type(t) == str for t in watched_events]
        ), f"{module_name}.watched_events entries must be of type string"
        assert all(
            [type(t) == str for t in produced_events]
        ), f"{module_name}.produced_events entries must be of type string"

        assert type(preloaded.get("deps_pip", [])) == list, f"{module_name}.deps_pipe must be of type list"
        assert type(preloaded.get("deps_apt", [])) == list, f"{module_name}.deps_apt must be of type list"
        assert type(preloaded.get("deps_shell", [])) == list, f"{module_name}.deps_shell must be of type list"
        assert type(preloaded.get("options", {})) == dict, f"{module_name}.options must be of type list"
        assert type(preloaded.get("options_desc", {})) == dict, f"{module_name}.options_desc must be of type list"
        # options must have descriptions
        assert set(preloaded.get("options", {})) == set(
            preloaded.get("options_desc", {})
        ), f"{module_name}.options do not match options_desc"
        # descriptions most not be blank
        assert all(
            o for o in preloaded.get("options_desc", {}).values()
        ), f"{module_name}.options_desc descriptions must not be blank"

    # setups
    futures = {}
    for module_name, module in scan2.modules.items():
        log.info(f"Testing {module_name}.setup()")
        future = scan2._thread_pool.submit_task(module.setup)
        futures[future] = module
    for future in helpers.as_completed(futures):
        module = futures[future]
        result = future.result()
        if type(result) == tuple:
            assert len(result) == 2, f"if tuple, {module.name}.setup() return value must have length of 2"
            status, msg = result
            assert status in (
                True,
                False,
                None,
            ), f"if tuple, the first element of {module.name}.setup()'s return value must be either True, False, or None"
            assert (
                type(msg) == str
            ), f"if tuple, the second element of {module.name}.setup()'s return value must be a message of type str"
        else:
            assert result in (
                True,
                False,
                None,
            ), f"{module.name}.setup() must return a status of either True, False, or None"
        if result == False:
            module.set_error_state()

    futures.clear()

    # handle_event / handle_batch
    futures = {}
    for module_name, module in scan2.modules.items():
        module.emit_event = lambda *args, **kwargs: None
        module._filter = lambda *args, **kwargs: True
        events_to_submit = [e for e in events.all if e.type in module.watched_events]
        if module.batch_size > 1:
            log.info(f"Testing {module_name}.handle_batch()")
            future = scan2._thread_pool.submit_task(module.handle_batch, *events_to_submit)
            futures[future] = module
        else:
            for e in events_to_submit:
                log.info(f"Testing {module_name}.handle_event()")
                future = scan2._thread_pool.submit_task(module.handle_event, e)
                futures[future] = module
    for future in helpers.as_completed(futures):
        try:
            assert future.result() == None
        except Exception as e:
            import traceback

            module = futures[future]
            assert module.errored == True, f'Error in module "{module}": {e}\n{traceback.format_exc()}'
    futures.clear()

    # finishes
    futures = {}
    for module_name, module in scan2.modules.items():
        log.info(f"Testing {module_name}.finish()")
        future = scan2._thread_pool.submit_task(module.finish)
        futures[future] = module
    for future in helpers.as_completed(futures):
        assert future.result() == None
    futures.clear()

    # cleanups
    futures = {}
    for module_name, module in scan2.modules.items():
        log.info(f"Testing {module_name}.cleanup()")
        future = scan2._thread_pool.submit_task(module.cleanup)
        futures[future] = module
    for future in helpers.as_completed(futures):
        assert future.result() == None
    futures.clear()

    # event filters
    for module_name, module in scan2.modules.items():
        log.info(f"Testing {module_name}.filter_event()")
        assert module.filter_event(events.emoji) in (True, False)


def test_config(bbot_config):
    from bbot.scanner.scanner import Scanner

    scan1 = Scanner("127.0.0.1", modules=["ipneighbor"], config=bbot_config)
    scan1.load_modules()
    assert scan1.config.plumbus == "asdf"
    assert scan1.modules["ipneighbor"].config.test_option == "ipneighbor"
    assert scan1.modules["human"].config.test_option == "human"
    assert scan1.modules["speculate"].config.test_option == "speculate"


def test_target(neuter_ansible, patch_requests, patch_commands, bbot_config):
    from bbot.scanner.scanner import Scanner

    scan1 = Scanner("api.publicapis.org", "8.8.8.8/30", "2001:4860:4860::8888/126", config=bbot_config)
    scan2 = Scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125", config=bbot_config)
    scan3 = Scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125", config=bbot_config)
    scan4 = Scanner("8.8.8.8/29", config=bbot_config)
    scan5 = Scanner(config=bbot_config)
    assert not scan5.target
    assert len(scan1.target) == 9
    assert len(scan4.target) == 8
    assert "8.8.8.9" in scan1.target
    assert "8.8.8.12" not in scan1.target
    assert "8.8.8.8/31" in scan1.target
    assert "8.8.8.8/30" in scan1.target
    assert "8.8.8.8/29" not in scan1.target
    assert "2001:4860:4860::8889" in scan1.target
    assert "2001:4860:4860::888c" not in scan1.target
    assert "www.api.publicapis.org" in scan1.target
    assert "api.publicapis.org" in scan1.target
    assert "publicapis.org" not in scan1.target
    assert "bob@www.api.publicapis.org" in scan1.target
    assert "https://www.api.publicapis.org" in scan1.target
    assert "www.api.publicapis.org:80" in scan1.target
    assert scan1.make_event("https://[2001:4860:4860::8888]:80", dummy=True) in scan1.target
    assert scan1.make_event("[2001:4860:4860::8888]:80", "OPEN_TCP_PORT", dummy=True) in scan1.target
    assert scan1.make_event("[2001:4860:4860::888c]:80", "OPEN_TCP_PORT", dummy=True) not in scan1.target
    assert scan1.target in scan2.target
    assert scan2.target not in scan1.target
    assert scan3.target in scan2.target
    assert scan2.target == scan3.target
    assert scan4.target != scan1.target


def test_scan(neuter_ansible, patch_requests, patch_commands, events, bbot_config, helpers, neograph):
    from bbot.scanner.scanner import Scanner

    scan0 = Scanner("8.8.8.8/31", "evilcorp.com", blacklist=["8.8.8.8/28", "www.evilcorp.com"], config=bbot_config)
    assert scan0.whitelisted("8.8.8.8")
    assert scan0.whitelisted("8.8.8.9")
    assert scan0.blacklisted("8.8.8.15")
    assert not scan0.blacklisted("8.8.8.16")
    assert scan0.blacklisted("8.8.8.8/30")
    assert not scan0.blacklisted("8.8.8.8/27")
    assert not scan0.in_scope("8.8.8.8")
    assert scan0.whitelisted("api.evilcorp.com")
    assert scan0.whitelisted("www.evilcorp.com")
    assert not scan0.blacklisted("api.evilcorp.com")
    assert scan0.blacklisted("asdf.www.evilcorp.com")
    assert scan0.in_scope("test.api.evilcorp.com")
    assert not scan0.in_scope("test.www.evilcorp.com")
    assert not scan0.in_scope("www.evilcorp.co.uk")

    scan1 = Scanner("8.8.8.8", whitelist=["8.8.4.4"], config=bbot_config)
    assert not scan1.blacklisted("8.8.8.8")
    assert not scan1.blacklisted("8.8.4.4")
    assert not scan1.whitelisted("8.8.8.8")
    assert scan1.whitelisted("8.8.4.4")
    assert scan1.in_scope("8.8.4.4")
    assert not scan1.in_scope("8.8.8.8")

    scan2 = Scanner("8.8.8.8", config=bbot_config)
    assert not scan2.blacklisted("8.8.8.8")
    assert not scan2.blacklisted("8.8.4.4")
    assert scan2.whitelisted("8.8.8.8")
    assert not scan2.whitelisted("8.8.4.4")
    assert scan2.in_scope("8.8.8.8")
    assert not scan2.in_scope("8.8.4.4")

    scan3 = Scanner(
        "127.0.0.0/30",
        "127.0.0.2:8443",
        "https://localhost",
        "[::1]:80",
        "http://[::1]:8080",
        modules=["ipneighbor"],
        output_modules=list(available_output_modules),
        config=bbot_config,
        blacklist=["http://127.0.0.3:8000/asdf"],
        whitelist=["127.0.0.0/29"],
    )
    assert "targets" in scan3.json
    assert "127.0.0.3" in scan3.target
    assert "127.0.0.4" not in scan3.target
    assert "127.0.0.4" in scan3.whitelist
    assert scan3.whitelisted("127.0.0.4")
    assert "127.0.0.3" in scan3.blacklist
    assert scan3.blacklisted("127.0.0.3")
    assert scan3.in_scope("127.0.0.1")
    assert not scan3.in_scope("127.0.0.3")
    scan3.start()


def test_threadpool():
    from concurrent.futures import ThreadPoolExecutor
    from bbot.core.helpers.threadpool import ThreadPoolWrapper, NamedLock, as_completed

    with ThreadPoolExecutor(max_workers=3) as executor:
        pool = ThreadPoolWrapper(executor)
        add_one = lambda x: x + 1
        futures = [pool.submit_task(add_one, y) for y in [0, 1, 2, 3, 4]]
        results = []
        for f in as_completed(futures):
            results.append(f.result())
        assert tuple(sorted(results)) == (1, 2, 3, 4, 5)

    nl = NamedLock(max_size=5)
    for i in range(50):
        nl.get_lock(str(i))
    assert len(nl._cache) == 5
    assert tuple(nl._cache.keys()) == tuple(hash(str(x)) for x in [45, 46, 47, 48, 49])


def test_agent(agent):
    agent.start()
    agent.on_error(agent.ws, "test")
    agent.on_close(agent.ws, "test", "test")
    agent.on_open(agent.ws)
    agent.on_message(
        agent.ws,
        '{"conversation": "90196cc1-299f-4555-82a0-bc22a4247590", "command": "start_scan", "arguments": {"scan_id": "90196cc1-299f-4555-82a0-bc22a4247590", "targets": ["www.blacklanternsecurity.com"], "modules": ["ipneighbor"], "output_modules": ["human"]}}',
    )
    sleep(0.5)
    agent.scan_status()
    agent.stop_scan()


def test_cli(monkeypatch, bbot_config):

    from bbot import cli

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(cli, "config", bbot_config)
    monkeypatch.setattr(sys, "argv", ["bbot", "-y", "--current-config", "-t", "127.0.0.1", "-m", "ipneighbor"])
    cli.main()

    home_dir = Path(bbot_config["home"])
    monkeypatch.setattr(
        sys,
        "argv",
        ["bbot", "-y", "-t", "localhost", "-m", "ipneighbor", "-om", "human", "csv", "json", "-n", "test_scan"],
    )
    cli.main()
    scan_home = home_dir / "scans" / "test_scan"
    assert (scan_home / "wordcloud.tsv").is_file()
    assert (scan_home / "output.txt").is_file()
    assert (scan_home / "output.csv").is_file()
    assert (scan_home / "output.json").is_file()
    with open(scan_home / "output.csv") as f:
        lines = f.readlines()
        assert lines[0] == "Event type,Event data,Source Module,Scope Distance,Event Tags\n"
        assert len(lines) > 1


def test_depsinstaller(monkeypatch, neuter_ansible, bbot_config):
    # un-neuter ansible
    from bbot.core.helpers.depsinstaller import installer

    run, ensure_root = neuter_ansible
    ensure_root = installer.DepsInstaller.ensure_root
    monkeypatch.setattr(installer, "run", run)
    monkeypatch.setattr(installer.DepsInstaller, "ensure_root", ensure_root)

    from bbot.scanner.scanner import Scanner

    scan = Scanner(
        "127.0.0.1",
        modules=["dnsresolve"],
        config=bbot_config,
    )

    # test shell
    test_file = Path("/tmp/test_file")
    test_file.unlink(missing_ok=True)
    scan.helpers.depsinstaller.shell(module="plumbus", commands=[f"touch {test_file}"])
    assert test_file.is_file()
    test_file.unlink(missing_ok=True)

    # test tasks
    scan.helpers.depsinstaller.tasks(
        module="plumbus",
        tasks=[{"name": "test task execution", "ansible.builtin.shell": {"cmd": f"touch {test_file}"}}],
    )
    assert test_file.is_file()
    test_file.unlink(missing_ok=True)



import os
import sys
from pathlib import Path
from omegaconf import OmegaConf

from . import files, args, environ
from ..errors import ConfigLoadError
from ...modules import module_loader
from ..helpers.misc import mkdir, error_and_exit, filter_dict, clean_dict

try:
    config = OmegaConf.merge(
        # first, pull module defaults
        OmegaConf.create(
            {
                "modules": module_loader.configs(type="scan"),
                "output_modules": module_loader.configs(type="output"),
                "internal_modules": module_loader.configs(type="internal"),
            }
        ),
        # then look in .yaml files
        files.get_config(),
        # finally, pull from CLI arguments
        args.get_config(),
    )
except ConfigLoadError as e:
    error_and_exit(e)

# ensure bbot_home
if not "home" in config:
    config["home"] = "~/.bbot"
home = Path(config["home"]).expanduser().resolve()
config["home"] = str(home)

# ensure bbot.yml
if not files.config_filename.exists():
    print(f"[INFO] Creating BBOT config at {files.config_filename}")
    no_secrets_config = OmegaConf.to_object(config)
    no_secrets_config = clean_dict(no_secrets_config, "api_key", "username", "password", "token", fuzzy=True)
    OmegaConf.save(config=OmegaConf.create(no_secrets_config), f=str(files.config_filename))

# ensure secrets.yml
if not files.secrets_filename.exists():
    print(f"[INFO] Creating BBOT secrets at {files.secrets_filename}")
    secrets_only_config = OmegaConf.to_object(config)
    secrets_only_config = filter_dict(secrets_only_config, "api_key", "username", "password", "token", fuzzy=True)
    OmegaConf.save(config=OmegaConf.create(secrets_only_config), f=str(files.secrets_filename))

# if we're running in a virtual environment, make sure to include its /bin in PATH
if sys.prefix != sys.base_prefix:
    bin_dir = str(Path(sys.prefix) / "bin")
    if bin_dir not in os.environ["PATH"].split(":"):
        os.environ["PATH"] = f'{bin_dir}:{os.environ.get("PATH", "")}'

# ensure bbot_tools
bbot_tools = home / "tools"
os.environ["BBOT_TOOLS"] = str(bbot_tools)
os.environ["PATH"] = f'{bbot_tools}:{os.environ.get("PATH", "")}'
# ensure bbot_cache
bbot_cache = home / "cache"
os.environ["BBOT_CACHE"] = str(bbot_cache)
# ensure bbot_temp
bbot_temp = home / "temp"
os.environ["BBOT_TEMP"] = str(bbot_temp)
# ensure bbot_lib
bbot_lib = home / "lib"
os.environ["BBOT_LIB"] = str(bbot_lib)

# exchange certain options between CLI args and config
if args.cli_options is not None:
    # deps
    config["retry_deps"] = args.cli_options.retry_deps
    config["force_deps"] = args.cli_options.force_deps
    config["no_deps"] = args.cli_options.no_deps
    config["ignore_failed_deps"] = args.cli_options.ignore_failed_deps
    # debug
    config["debug"] = args.cli_options.debug
    if args.cli_options.output_dir:
        config["output_dir"] = args.cli_options.output_dir

# copy config to environment
bbot_environ = environ.flatten_config(config)
os.environ.update(bbot_environ)

# handle HTTP proxy
http_proxy = config.get("http_proxy", "")
if http_proxy:
    os.environ["HTTP_PROXY"] = http_proxy
    os.environ["HTTPS_PROXY"] = http_proxy

# export LD_LIBRARY_PATH=${LD_LIBRARY_PATH}:~/.bbot/lib/
os.environ["LD_LIBRARY_PATH"] = ":".join(os.environ.get("LD_LIBRARY_PATH", "").split(":") + [str(bbot_lib)]).strip(":")

# replace environment variables in preloaded modules
module_loader.find_and_replace(**os.environ)

# ssl verification
import urllib3

urllib3.disable_warnings()
ssl_verify = config.get("ssl_verify", False)
if not ssl_verify:
    import requests
    import functools

    requests.adapters.BaseAdapter.send = functools.partialmethod(requests.adapters.BaseAdapter.send, verify=False)
    requests.adapters.HTTPAdapter.send = functools.partialmethod(requests.adapters.HTTPAdapter.send, verify=False)
    requests.Session.request = functools.partialmethod(requests.Session.request, verify=False)
    requests.request = functools.partial(requests.request, verify=False)



from .shodan_dns import shodan_dns


class c99(shodan_dns):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the C99 API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "c99.nl API key"}

    base_url = "https://api.c99.nl"

    def ping(self):
        url = f"{self.base_url}/randomnumber?key={self.api_key}&between=1,100&json"
        response = self.helpers.request(url)
        assert response.json()["success"] == True

    def query(self, query):
        url = f"{self.base_url}/subdomainfinder?key={self.api_key}&domain={self.helpers.quote(query)}&json"
        results = self.helpers.request(url)
        try:
            json = results.json()
            if json:
                subdomains = json.get("subdomains", [])
                if subdomains:
                    for s in subdomains:
                        subdomain = s.get("subdomain", "")
                        if subdomain:
                            yield subdomain
            else:
                self.debug(f'No results for "{query}"')
        except Exception:
            import traceback

            self.warning(f"Error retrieving c99.nl subdomains for {query}")
            self.debug(traceback.format_exc())



import re
import json
import logging
import dns.resolver
import dns.exception
from threading import Lock
from contextlib import suppress
from concurrent.futures import ThreadPoolExecutor

from .regexes import dns_name_regex
from bbot.core.errors import ValidationError
from .threadpool import ThreadPoolWrapper, NamedLock
from .misc import is_ip, domain_parents, parent_domain, rand_string

log = logging.getLogger("bbot.core.helpers.dns")


class DNSHelper:
    """
    For automatic wildcard detection, nameserver validation, etc.
    """

    def __init__(self, parent_helper):

        self.parent_helper = parent_helper
        self.resolver = dns.resolver.Resolver()
        self.timeout = self.parent_helper.config.get("dns_timeout", 10)
        self.abort_threshold = self.parent_helper.config.get("dns_abort_threshold", 5)
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout
        self._resolver_list = None

        self.wildcard_ignore = self.parent_helper.config.get("dns_wildcard_ignore", None)
        if not self.wildcard_ignore:
            self.wildcard_ignore = []
        self.wildcard_ignore = tuple([str(d).strip().lower() for d in self.wildcard_ignore])
        self.wildcard_tests = self.parent_helper.config.get("dns_wildcard_tests", 5)
        self._wildcard_cache = dict()
        # since wildcard detection takes some time, This is to prevent multiple
        # modules from kicking off wildcard detection for the same domain at the same time
        self._wildcard_lock = NamedLock()

        self._errors = dict()
        self._error_lock = Lock()

        # we need our own threadpool because using the shared one can lead to deadlocks
        max_workers = self.parent_helper.config.get("max_dns_threads", 100)
        executor = ThreadPoolExecutor(max_workers=max_workers)
        self._thread_pool = ThreadPoolWrapper(executor, max_workers=max_workers)

        self._debug = self.parent_helper.config.get("dns_debug", False)

        self._dummy_modules = dict()
        self._dummy_modules_lock = Lock()

        self._cache = self.parent_helper.CacheDict(max_size=10000)
        self._cache_lock = Lock()
        self._cache_locks = NamedLock()

        # copy the system's current resolvers to a text file for tool use
        resolvers = dns.resolver.Resolver().nameservers
        self.resolver_file = self.parent_helper.tempfile(resolvers, pipe=False)

        self.bad_ptr_regex = re.compile(r"(?:[0-9]{1,3}[-_\.]){3}[0-9]{1,3}")
        self.filter_bad_ptrs = self.parent_helper.config.get("dns_filter_ptrs", True)

    def resolve(self, query, **kwargs):
        """
        "1.2.3.4" --> {
            "evilcorp.com",
        }
        "evilcorp.com" --> {
            "1.2.3.4",
            "dead::beef"
        }
        """
        results = set()
        raw_results, errors = self.resolve_raw(query, **kwargs)
        for (rdtype, answers) in raw_results:
            for answer in answers:
                for _, t in self.extract_targets(answer):
                    results.add(t)
        return results

    def resolve_raw(self, query, **kwargs):
        kwargs["tcp"] = True
        if self.parent_helper.scan.stopping:
            return [], []
        query = str(query).strip()
        if is_ip(query):
            kwargs.pop("type", None)
            kwargs.pop("rdtype", None)
            results, errors = self._resolve_ip(query, **kwargs)
            return [("PTR", results)], [("PTR", e) for e in errors]
        else:
            results = []
            errors = []
            types = ["A", "AAAA"]
            kwargs.pop("rdtype", None)
            if "type" in kwargs:
                t = kwargs.pop("type")
                if isinstance(t, str):
                    if t.strip().lower() in ("any", "all", "*"):
                        types = ["A", "AAAA", "SRV", "MX", "NS", "SOA", "CNAME", "TXT"]
                    else:
                        types = [t.strip().upper()]
                elif any([isinstance(t, x) for x in (list, tuple)]):
                    types = [str(_).strip().upper() for _ in t]
            for t in types:
                if getattr(self.parent_helper.scan, "stopping", False) == True:
                    break
                r, e = self._resolve_hostname(query, rdtype=t, **kwargs)
                if r:
                    results.append((t, r))
                for error in e:
                    errors.append((t, error))

            return (results, errors)

    def _resolve_hostname(self, query, **kwargs):
        self.debug(f"Resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        parent = self.parent_helper.parent_domain(query)
        rdtype = kwargs.get("rdtype", "A")
        parent_hash = hash(f"{parent}:{rdtype}")
        error_count = self._errors.get(parent_hash, 0)
        if error_count >= self.abort_threshold:
            log.verbose(
                f'Aborting query "{query}" because failed {rdtype} queries for "{parent}" ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
            )
            return results, errors
        try:
            results = list(self._catch(self.resolver.resolve, query, **kwargs))
            with self._error_lock:
                if parent_hash in self._errors:
                    self._errors[parent_hash] = 0
        except (dns.resolver.NoNameservers, dns.exception.Timeout, dns.resolver.LifetimeTimeout) as e:
            with self._error_lock:
                try:
                    self._errors[parent_hash] += 1
                except KeyError:
                    self._errors[parent_hash] = 1
                log.verbose(
                    f'DNS error or timeout for {rdtype} query "{query}" ({self._errors[parent_hash]:,} so far): {e}'
                )
                errors.append(e)
        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results, errors

    def _resolve_ip(self, query, **kwargs):
        self.debug(f"Reverse-resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        try:
            return list(self._catch(self.resolver.resolve_address, query, **kwargs)), errors
        except dns.resolver.NoNameservers as e:
            self.debug(f"{e} (query={query}, kwargs={kwargs})")
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout) as e:
            errors.append(e)
        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results, errors

    def resolve_event(self, event):
        result = self._resolve_event(event)
        # if it's a wildcard, go again with _wildcard.{domain}
        if len(result) == 1:
            event = result[0]
            return self._resolve_event(event, check_wildcard=False)
        # else we're good
        else:
            return result

    def _resolve_event(self, event, check_wildcard=True):
        """
        Tag event with appropriate dns record types
        Optionally create child events from dns resolutions
        """
        event_tags = set()
        try:
            if not event.host or event.type in ("IP_RANGE",):
                return [], set(), False, False
            children = []
            event_host = str(event.host)
            # lock to ensure resolution of the same host doesn't start while we're working here
            with self._cache_locks.get_lock(event_host):

                event_whitelisted = False
                event_blacklisted = False

                # wildcard check first
                if check_wildcard:
                    event_is_wildcard, wildcard_parent = self.is_wildcard(event_host)
                    if event_is_wildcard and event.type in ("DNS_NAME",):
                        event.data = wildcard_parent
                        return (event,)
                else:
                    event_tags.add("wildcard")

                # try to get data from cache
                _event_tags, _event_whitelisted, _event_blacklisted = self.cache_get(event_host)
                event_tags.update(_event_tags)
                # if we found it, return it
                if _event_whitelisted is not None:
                    return children, event_tags, _event_whitelisted, _event_blacklisted

                # then resolve
                resolved_raw, errors = self.resolve_raw(event_host, type="any")
                if errors:
                    event_tags.add("dns-error")
                for rdtype, records in resolved_raw:
                    event_tags.add("resolved")
                    event_tags.add(f"{rdtype.lower()}_record")
                    rdtype = str(rdtype).upper()
                    # whitelisting and blacklist of IPs
                    if rdtype in ("A", "AAAA"):
                        for r in records:
                            for _, t in self.extract_targets(r):
                                with suppress(ValidationError):
                                    if self.parent_helper.scan.whitelisted(t):
                                        event_whitelisted = True
                                with suppress(ValidationError):
                                    if self.parent_helper.scan.blacklisted(t):
                                        event_blacklisted = True
                    for r in records:
                        for _, t in self.extract_targets(r):
                            if t:
                                if self.filter_bad_ptrs and rdtype in ("PTR") and self.bad_ptr_regex.search(t):
                                    self.debug(f"Filtering out bad PTR: {t}")
                                    continue
                                children.append((t, rdtype))
                if "resolved" not in event_tags:
                    event_tags.add("unresolved")
                self._cache[event_host] = (event_tags, event_whitelisted, event_blacklisted)
            return children, event_tags, event_whitelisted, event_blacklisted
        finally:
            event._resolved.set()

    def cache_get(self, host):
        try:
            return self._cache[host]
        except KeyError:
            return set(), None, None

    def resolve_batch(self, queries, **kwargs):
        """
        resolve_batch("www.evilcorp.com", "evilcorp.com") --> [
            ("www.evilcorp.com", {"1.1.1.1"}),
            ("evilcorp.com", {"2.2.2.2"})
        ]
        """
        futures = dict()
        for query in queries:
            future = self._thread_pool.submit_task(self._catch_keyboardinterrupt, self.resolve, query, **kwargs)
            futures[future] = query
        for future in self.parent_helper.as_completed(futures):
            query = futures[future]
            yield (query, future.result())

    def extract_targets(self, record):
        """
        Extract whatever hostnames/IPs a DNS records points to
        """
        results = set()
        rdtype = str(record.rdtype.name).upper()
        if rdtype in ("A", "AAAA", "NS", "CNAME", "PTR"):
            results.add((rdtype, self._clean_dns_record(record)))
        elif rdtype == "SOA":
            results.add((rdtype, self._clean_dns_record(record.mname)))
        elif rdtype == "MX":
            results.add((rdtype, self._clean_dns_record(record.exchange)))
        elif rdtype == "SRV":
            results.add((rdtype, self._clean_dns_record(record.target)))
        elif rdtype == "TXT":
            for s in record.strings:
                s = self.parent_helper.smart_decode(s)
                for match in dns_name_regex.finditer(s):
                    start, end = match.span()
                    host = s[start:end]
                    results.add((rdtype, host))
        else:
            log.warning(f'Unknown DNS record type "{rdtype}"')
        return results

    @staticmethod
    def _clean_dns_record(record):
        return str(record.to_text()).lower().rstrip(".")

    def get_valid_resolvers(self, min_reliability=0.99):
        nameservers = set()
        nameservers_url = "https://public-dns.info/nameserver/nameservers.json"
        nameservers_file = self.parent_helper.download(nameservers_url, cache_hrs=72)
        if nameservers_file is None:
            log.warning(f"Failed to download nameservers from {nameservers_url}")
            return set()
        nameservers_json = []
        try:
            nameservers_json = json.loads(open(nameservers_file).read())
        except Exception as e:
            log.warning(f"Failed to load nameserver list from {nameservers_file}: {e}")
            nameservers_file.unlink()
        for entry in nameservers_json:
            try:
                ip = str(entry.get("ip", "")).strip()
            except Exception:
                continue
            try:
                reliability = float(entry.get("reliability", 0))
            except ValueError:
                continue
            if reliability >= min_reliability and is_ip(ip, version=4):
                nameservers.add(ip)
        log.verbose(f"Loaded {len(nameservers):,} nameservers from {nameservers_url}")
        resolver_list = self.verify_nameservers(nameservers)
        return resolver_list

    @property
    def resolvers(self):
        """
        Returns set() of valid DNS servers from public-dns.info
        """
        if self._resolver_list is None:
            file_content = self.parent_helper.cache_get("resolver_list")
            if file_content is not None:
                self._resolver_list = set([l for l in file_content.splitlines() if l])
            if not self._resolver_list:
                log.info(f"Fetching and validating public DNS servers, this may take a few minutes")
                resolvers = self.get_valid_resolvers()
                if resolvers:
                    self._resolver_list = resolvers
                    self.parent_helper.cache_put("resolver_list", "\n".join(self._resolver_list))
                else:
                    return set()
        return self._resolver_list

    @property
    def mass_resolver_file(self):
        self.resolvers
        return self.parent_helper.cache_filename("resolver_list")

    def verify_nameservers(self, nameservers, timeout=2):
        """Check each resolver to make sure it can actually resolve DNS names

        Args:
            nameservers (list): nameservers to verify
            timeout (int): timeout for dns query
        """
        log.verbose(f"Verifying {len(nameservers):,} nameservers")
        futures = [
            self._thread_pool.submit_task(self._catch_keyboardinterrupt, self.verify_nameserver, n)
            for n in nameservers
        ]

        valid_nameservers = set()
        for future in self.parent_helper.as_completed(futures):
            nameserver, error = future.result()
            if error is None:
                self.debug(f'Nameserver "{nameserver}" is valid')
                valid_nameservers.add(nameserver)
            else:
                self.debug(str(error))
        log.verbose(f"Verified {len(valid_nameservers):,}/{len(nameservers):,} nameservers")

        return valid_nameservers

    def verify_nameserver(self, nameserver, timeout=2):
        """Validate a nameserver by making a sample query and a garbage query

        Args:
            nameserver (str): nameserver to verify
            timeout (int): timeout for dns query
        """
        self.debug(f'Verifying nameserver "{nameserver}"')
        error = None

        resolver = dns.resolver.Resolver()
        resolver.timeout = timeout
        resolver.lifetime = timeout
        resolver.nameservers = [nameserver]

        # first, make sure it can resolve a valid hostname
        try:
            a_results = [str(r) for r in list(resolver.resolve("dns.google", "A"))]
            aaaa_results = [str(r) for r in list(resolver.resolve("dns.google", "AAAA"))]
            if not ("2001:4860:4860::8888" in aaaa_results and "8.8.8.8" in a_results):
                error = f"Nameserver {nameserver} failed to resolve basic query"
        except Exception:
            error = f"Nameserver {nameserver} failed to resolve basic query within {timeout} seconds"

        # then, make sure it isn't feeding us garbage data
        randhost = f"www-m.{rand_string(9)}.{rand_string(10)}.com"
        if error is None:
            try:
                a_results = list(resolver.resolve(randhost, "A"))
                error = f"Nameserver {nameserver} returned garbage data"
            except dns.exception.DNSException:
                pass
                # Garbage query to nameserver failed successfully ;)
        if error is None:
            try:
                a_results = list(resolver.resolve(randhost, "AAAA"))
                error = f"Nameserver {nameserver} returned garbage data"
            except dns.exception.DNSException:
                pass
                # Garbage query to nameserver failed successfully ;)

        return nameserver, error

    def _catch(self, callback, *args, **kwargs):
        try:
            return callback(*args, **kwargs)
        except dns.resolver.NoNameservers:
            raise
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout):
            log.debug(f"DNS query with args={args}, kwargs={kwargs} timed out after {self.timeout} seconds")
            raise
        except dns.exception.DNSException as e:
            self.debug(f"{e} (args={args}, kwargs={kwargs})")
        except Exception:
            log.warning(f"Error in {callback.__qualname__}() with args={args}, kwargs={kwargs}")
        return list()

    def is_wildcard(self, query):
        if is_ip(query) or not "." in query:
            return False, query
        for d in self.wildcard_ignore:
            if self.parent_helper.host_in_host(query, d):
                return False, query
        if "_wildcard" in query.split("."):
            return True, query
        hosts = list(domain_parents(query, include_self=True))[:-1]
        for host in hosts[::-1]:
            is_wildcard, parent = self._is_wildcard(host)
            if is_wildcard:
                return True, f"_wildcard.{parent}"
        return False, query

    def _is_wildcard(self, query):
        parent = parent_domain(query)
        parent_hash = hash(parent)

        # try to return from cache
        with suppress(KeyError):
            return self._wildcard_cache[parent_hash], parent

        with self._wildcard_lock.get_lock(parent):

            # resolve the base query
            orig_results = self.resolve(query)
            is_wildcard = False

            futures = []
            # resolve a bunch of random subdomains of the same parent
            for _ in range(self.wildcard_tests):
                rand_query = f"{rand_string(length=10)}.{parent}"
                future = self._thread_pool.submit_task(self._catch_keyboardinterrupt, self.resolve, rand_query)
                futures.append(future)

            # put all the IPs from the random subdomains in one bucket
            wildcard_ips = set()
            for future in self.parent_helper.as_completed(futures):
                ips = future.result()
                if ips:
                    wildcard_ips.update(ips)

            # if all of the original results are in the random bucket
            if orig_results and wildcard_ips and all([ip in wildcard_ips for ip in orig_results]):
                # then ladies and gentlemen we have a wildcard
                is_wildcard = True

            self._wildcard_cache.update({parent_hash: is_wildcard})
            if is_wildcard:
                log.verbose(f"Encountered domain with wildcard DNS: {parent}")
            return is_wildcard, parent

    def _catch_keyboardinterrupt(self, callback, *args, **kwargs):
        try:
            return callback(*args, **kwargs)
        except Exception as e:
            import traceback

            log.error(f"Error in {callback.__qualname__}(): {e}")
            log.debug(traceback.format_exc())
        except KeyboardInterrupt:
            if self.parent_helper.scan:
                self.parent_helper.scan.stop()

    def debug(self, *args, **kwargs):
        if self._debug:
            log.debug(*args, **kwargs)

    def _get_dummy_module(self, name):
        with self._dummy_modules_lock:
            try:
                dummy_module = self._dummy_modules[name]
            except KeyError:
                dummy_module = self.parent_helper._make_dummy_module(name=name, _type="DNS")
                self._dummy_modules[name] = dummy_module
        return dummy_module



from pathlib import Path
from shutil import copyfile
from omegaconf import OmegaConf

from ..helpers.misc import mkdir
from ..errors import ConfigLoadError

config_dir = (Path.home() / ".config" / "bbot").resolve()
defaults_filename = (Path(__file__).parent.parent.parent / "defaults.yml").resolve()
defaults_destination = config_dir / "defaults.yml"
mkdir(config_dir)
copyfile(defaults_filename, defaults_destination)
config_filename = (config_dir / "bbot.yml").resolve()
secrets_filename = (config_dir / "secrets.yml").resolve()


def _get_config(filename):
    filename = Path(filename).resolve()
    try:
        return OmegaConf.load(str(filename))
    except Exception as e:
        if filename.exists():
            raise ConfigLoadError(f"Error parsing config at {filename}:\n\n{e}")
        return OmegaConf.create()


def get_config():

    return OmegaConf.merge(
        _get_config(defaults_filename),
        _get_config(config_filename),
        _get_config(secrets_filename),
    )



import pytest
import urllib3
import requests
import tldextract
from pathlib import Path
from omegaconf import OmegaConf

# make the necessary web requests before nuking them to high heaven
example_url = "https://api.publicapis.org/health"
http = urllib3.PoolManager()
urllib_response = http.request("GET", example_url)
requests_response = requests.get(example_url)
tldextract.extract("www.evilcorp.com")


@pytest.fixture
def neograph(monkeypatch, helpers):
    helpers.depsinstaller.pip_install(["py2neo"])

    class NeoGraph:
        def __init__(self, *args, **kwargs):
            pass

        def merge(self, *args, **kwargs):
            return True

    import py2neo

    monkeypatch.setattr(py2neo, "Graph", NeoGraph)
    from bbot.db.neo4j import Neo4j

    return Neo4j(uri="bolt://127.0.0.1:1111")


@pytest.fixture
def neuter_ansible(monkeypatch):
    from ansible_runner.interface import run

    class AnsibleRunnerResult:
        status = "successful"
        rc = 0
        events = []

    def ansible_run(*args, **kwargs):
        module = kwargs.get("module", "")
        if module != "pip":
            return AnsibleRunnerResult()
        else:
            return run(*args, **kwargs)

    from bbot.core.helpers.depsinstaller import installer

    ensure_root = installer.DepsInstaller.ensure_root

    monkeypatch.setattr(installer, "run", ansible_run)
    monkeypatch.setattr(installer.DepsInstaller, "ensure_root", lambda *args, **kwargs: None)

    return run, ensure_root


@pytest.fixture
def bbot_config():
    from bbot import config as default_config

    test_config = OmegaConf.load(Path(__file__).parent / "test.conf")
    config = OmegaConf.merge(default_config, test_config)
    return config


@pytest.fixture
def scan(neuter_ansible, patch_requests, patch_commands, bbot_config):
    from bbot.scanner import Scanner

    bbot_scan = Scanner("127.0.0.1", modules=["ipneighbor"], config=bbot_config)
    bbot_scan.status = "RUNNING"
    return bbot_scan


@pytest.fixture
def helpers(scan):
    return scan.helpers


httpx_response = {
    "timestamp": "2022-06-29T09:56:19.927240577-04:00",
    "request": "GET / HTTP/1.1\r\nHost: example.com\r\nUser-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.35 Safari/537.36\r\nAccept-Charset: utf-8\r\nAccept-Encoding: gzip\r\n\r\n",
    "response-header": 'HTTP/1.1 200 OK\r\nConnection: close\r\nAccept-Ranges: bytes\r\nAge: 557710\r\nCache-Control: max-age=604800\r\nContent-Type: text/html; charset=UTF-8\r\nDate: Wed, 29 Jun 2022 13:56:16 GMT\r\nEtag: "3147526947"\r\nExpires: Wed, 06 Jul 2022 13:56:16 GMT\r\nLast-Modified: Thu, 17 Oct 2019 07:18:26 GMT\r\nServer: ECS (agb/A438)\r\nVary: Accept-Encoding\r\nX-Cache: HIT\r\n\r\n',
    "scheme": "http",
    "port": "80",
    "path": "/",
    "url": "http://example.com:80",
    "input": "http://example.com",
    "title": "Example Domain",
    "webserver": "ECS (agb/A438)",
    "response-body": '<!doctype html>\n<html>\n<head>\n    <title>Example Domain</title>\n\n    <meta charset="utf-8" />\n    <meta http-equiv="Content-type" content="text/html; charset=utf-8" />\n    <meta name="viewport" content="width=device-width, initial-scale=1" />\n    <style type="text/css">\n    body {\n        background-color: #f0f0f2;\n        margin: 0;\n        padding: 0;\n        font-family: -apple-system, system-ui, BlinkMacSystemFont, "Segoe UI", "Open Sans", "Helvetica Neue", Helvetica, Arial, sans-serif;\n        \n    }\n    div {\n        width: 600px;\n        margin: 5em auto;\n        padding: 2em;\n        background-color: #fdfdff;\n        border-radius: 0.5em;\n        box-shadow: 2px 3px 7px 2px rgba(0,0,0,0.02);\n    }\n    a:link, a:visited {\n        color: #38488f;\n        text-decoration: none;\n    }\n    @media (max-width: 700px) {\n        div {\n            margin: 0 auto;\n            width: auto;\n        }\n    }\n    </style>    \n</head>\n\n<body>\n<div>\n    <h1>Example Domain</h1>\n    <p>This domain is for use in illustrative examples in documents. You may use this\n    domain in literature without prior coordination or asking for permission.</p>\n    <p><a href="https://www.iana.org/domains/example">More information...</a></p>\n</div>\n</body>\n</html>\n',
    "content-type": "text/html",
    "method": "GET",
    "host": "93.184.216.34",
    "content-length": 1256,
    "status-code": 200,
    "response-time": "95.343985ms",
    "failed": False,
    "hashes": {
        "body-md5": "84238dfc8092e5d9c0dac8ef93371a07",
        "body-mmh3": "-1139337416",
        "body-sha256": "ea8fac7c65fb589b0d53560f5251f74f9e9b243478dcb6b3ea79b5e36449c8d9",
        "body-simhash": "9899951357530060719",
        "header-md5": "99b650ea40a9e95550d7540996b67b60",
        "header-mmh3": "1831947040",
        "header-sha256": "eecbd4d9798c44295df0c5f2beebd939e7e51d9c6c16842dd73be83273f406bd",
        "header-simhash": "15614709017155964779",
    },
    "lines": 47,
    "words": 298,
}


@pytest.fixture
def events(scan):
    class bbot_events:
        localhost = scan.make_event("127.0.0.1", dummy=True)
        ipv4 = scan.make_event("8.8.8.8", dummy=True)
        netv4 = scan.make_event("8.8.8.8/30", dummy=True)
        ipv6 = scan.make_event("2001:4860:4860::8888", dummy=True)
        netv6 = scan.make_event("2001:4860:4860::8888/126", dummy=True)
        domain = scan.make_event("publicAPIs.org", dummy=True)
        subdomain = scan.make_event("api.publicAPIs.org", dummy=True)
        email = scan.make_event("bob@evilcorp.co.uk", "EMAIL_ADDRESS", dummy=True)
        open_port = scan.make_event("api.publicAPIs.org:443", dummy=True)
        protocol = scan.make_event({"host": "api.publicAPIs.org:443", "protocol": "HTTP"}, "PROTOCOL", dummy=True)
        ipv4_open_port = scan.make_event("8.8.8.8:443", dummy=True)
        ipv6_open_port = scan.make_event("[2001:4860:4860::8888]:443", "OPEN_TCP_PORT", dummy=True)
        url_unverified = scan.make_event("https://api.publicAPIs.org:443/hellofriend", dummy=True)
        ipv4_url_unverified = scan.make_event("https://8.8.8.8:443/hellofriend", dummy=True)
        ipv6_url_unverified = scan.make_event("https://[2001:4860:4860::8888]:443/hellofriend", dummy=True)
        url = scan.make_event("https://api.publicAPIs.org:443/hellofriend", "URL", dummy=True)
        ipv4_url = scan.make_event("https://8.8.8.8:443/hellofriend", "URL", dummy=True)
        ipv6_url = scan.make_event("https://[2001:4860:4860::8888]:443/hellofriend", "URL", dummy=True)
        url_hint = scan.make_event("https://api.publicAPIs.org:443/hello.ash", "URL_HINT", dummy=True)
        vulnerability = scan.make_event(
            {"host": "evilcorp.com", "severity": "INFO", "description": "asdf"}, "VULNERABILITY", dummy=True
        )
        finding = scan.make_event({"host": "evilcorp.com", "description": "asdf"}, "FINDING", dummy=True)
        vhost = scan.make_event({"host": "evilcorp.com", "vhost": "www.evilcorp.com"}, "VHOST", dummy=True)
        http_response = scan.make_event(httpx_response, "HTTP_RESPONSE", dummy=True)
        emoji = scan.make_event("💩", "WHERE_IS_YOUR_GOD_NOW", dummy=True)

    bbot_events.all = [  # noqa: F841
        bbot_events.localhost,
        bbot_events.ipv4,
        bbot_events.netv4,
        bbot_events.ipv6,
        bbot_events.netv6,
        bbot_events.domain,
        bbot_events.subdomain,
        bbot_events.email,
        bbot_events.open_port,
        bbot_events.protocol,
        bbot_events.ipv4_open_port,
        bbot_events.ipv6_open_port,
        bbot_events.url_unverified,
        bbot_events.ipv4_url_unverified,
        bbot_events.ipv6_url_unverified,
        bbot_events.url,
        bbot_events.ipv4_url,
        bbot_events.ipv6_url,
        bbot_events.url_hint,
        bbot_events.vulnerability,
        bbot_events.finding,
        bbot_events.vhost,
        bbot_events.http_response,
        bbot_events.emoji,
    ]

    for e in bbot_events.all:
        e.make_in_scope()

    return bbot_events


@pytest.fixture
def patch_requests(monkeypatch):
    from bbot.core.helpers.web import request, download

    monkeypatch.setattr("urllib3.connectionpool.HTTPConnectionPool.urlopen", lambda *args, **kwargs: urllib_response)
    monkeypatch.setattr("urllib3.poolmanager.PoolManager.urlopen", lambda *args, **kwargs: urllib_response)
    monkeypatch.setattr("requests.adapters.HTTPAdapter.send", lambda *args, **kwargs: requests_response)
    monkeypatch.setattr("bbot.core.helpers.web.request", lambda *args, **kwargs: requests_response)
    current_dir = Path(__file__).resolve().parent
    downloaded_file = current_dir / "test_output.json"
    monkeypatch.setattr("bbot.core.helpers.web.download", lambda *args, **kwargs: downloaded_file)
    return request, download


@pytest.fixture
def patch_commands(monkeypatch):
    import subprocess

    sample_output = [
        # massdns
        """{"name":"www.blacklanternsecurity.com.","type":"A","class":"IN","status":"NOERROR","rx_ts":1659985004071981831,"data":{"answers":[{"ttl":3580,"type":"CNAME","class":"IN","name":"www.blacklanternsecurity.com.","data":"blacklanternsecurity.github.io."},{"ttl":3580,"type":"A","class":"IN","name":"blacklanternsecurity.github.io.","data":"185.199.108.153"},{"ttl":3580,"type":"A","class":"IN","name":"blacklanternsecurity.github.io.","data":"185.199.109.153"},{"ttl":3580,"type":"A","class":"IN","name":"blacklanternsecurity.github.io.","data":"185.199.110.153"},{"ttl":3580,"type":"A","class":"IN","name":"blacklanternsecurity.github.io.","data":"185.199.111.153"}]},"flags":["rd","ra"],"resolver":"8.8.8.8:53"}""",
        # httpx
        """{"timestamp":"2022-04-15T17:08:29.436778586-04:00","request":"GET /health HTTP/1.1\\r\\nHost: api.publicapis.org\\r\\nUser-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.131 Safari/537.36\\r\\nAccept-Charset: utf-8\\r\\nAccept-Encoding: gzip\\r\\n\\r\\n","response-header":"HTTP/1.1 200 OK\\r\\nConnection: close\\r\\nContent-Length: 15\\r\\nContent-Type: text/plain; charset=utf-8\\r\\nDate: Fri, 15 Apr 2022 21:08:29 GMT\\r\\nServer: Caddy\\r\\nX-Rate-Limit-Duration: 1\\r\\nX-Rate-Limit-Limit: 10.00\\r\\nX-Rate-Limit-Request-Forwarded-For: 50.240.76.25\\r\\nX-Rate-Limit-Request-Remote-Addr: 172.17.0.1:32910\\r\\n\\r\\n","scheme":"https","port":"443","path":"/health","body-sha256":"6c63d4b385b07fe0e09a8a1f95b826e8a7d0401dfd12d649fe7c64b8a785023e","header-sha256":"161187846622dc97219392ab70195f4a477457e55dadf4b39f1b6c734e396120","url":"https://api.publicapis.org:443/health","input":"https://api.publicapis.org/health","webserver":"Caddy","response-body":"{\\"alive\\": true}","content-type":"text/plain","method":"GET","host":"138.197.231.124","content-length":15,"status-code":200,"response-time":"412.587433ms","failed":false,"lines":1,"words":2}""",
        # nuclei
        """{"template":"technologies/tech-detect.yaml","template-url":"https://github.com/projectdiscovery/nuclei-templates/blob/master/technologies/tech-detect.yaml","template-id":"tech-detect","info":{"name":"Wappalyzer Technology Detection","author":["hakluke"],"tags":["tech"],"reference":null,"severity":"info"},"matcher-name":"caddy","type":"http","host":"https://api.publicapis.org/health","matched-at":"https://api.publicapis.org:443/health","ip":"138.197.231.124","timestamp":"2022-04-15T17:09:01.021589723-04:00","curl-command":"curl -X 'GET' -d '' -H 'Accept: */*' -H 'Accept-Language: en' -H 'User-Agent: Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2226.0 Safari/537.36' 'https://api.publicapis.org/health'","matcher-status":true,"matched-line":null}""",
        # naabu
        """{"ip":"8.8.8.8","port":443,"timestamp":"2022-08-03T16:01:15.684442081Z"}"""
        # ffuf
        """{"input":{"FUZZ":"L2luZGV4Lmh0bWw="},"position":1,"status":200,"length":1256,"words":298,"lines":47,"content-type":"text/html;charset=UTF-8","redirectlocation":"","url":"http://example.com:80//index.html","duration":101243249,"resultfile":"","host":"example.com:80"}""",
        "https://api.publicapis.org:443/health",
        # open port
        "api.publicapis.org:443",
        # host
        "api.publicapis.org",
        # url
        "https://8.8.8.8",
    ]

    def run(*args, **kwargs):
        text = kwargs.get("text", True)
        return subprocess.run(["echo", "\n".join(sample_output)], text=text, stdout=subprocess.PIPE)

    def run_live(*args, **kwargs):
        for line in sample_output:
            yield line

    from bbot.core.helpers.command import run as old_run, run_live as old_run_live

    monkeypatch.setattr("bbot.core.helpers.command.run", run)
    monkeypatch.setattr("bbot.core.helpers.command.run_live", run_live)

    return old_run, old_run_live


@pytest.fixture
def agent(monkeypatch):
    class WebSocketApp:
        def __init__(*args, **kwargs):
            return

        def send(self, message):
            assert type(message) == str

        def run_forever(*args, **kwargs):
            return False

        def close(self):
            return

    from bbot import agent
    from bbot.modules.output.websocket import Websocket

    monkeypatch.setattr(Websocket, "send", lambda *args, **kwargs: True)

    test_agent = agent.Agent({"agent_url": "test", "agent_token": "test"})
    test_agent.setup()
    monkeypatch.setattr(test_agent, "ws", WebSocketApp())
    return test_agent



from bbot.modules.base import BaseModule


class iis_shortnames(BaseModule):

    watched_events = ["URL"]
    produced_events = ["URL_HINT"]
    flags = ["active", "safe"]
    meta = {"description": "Check for IIS shortname vulnerability"}
    options = {"detect_only": True, "threads": 8}
    options_desc = {
        "detect_only": "Only detect the vulnerability and do not run the shortname scanner",
        "threads": "the number of threads to run concurrently when executing the IIS shortname scanner",
    }
    in_scope_only = True

    deps_ansible = [
        {
            "name": "Install Java JRE (Debian)",
            "become": True,
            "package": {"name": "default-jre", "state": "latest"},
            "when": """ansible_facts['os_family'] == 'Debian'""",
        },
        {
            "name": "Install Java JRE (RedHat)",
            "become": True,
            "package": {"name": "java-latest-openjdk", "state": "latest"},
            "when": """ansible_facts['os_family'] == 'RedHat'""",
        },
        {
            "name": "Install Java JRE (Archlinux)",
            "package": {"name": "jre-openjdk", "state": "present"},
            "become": True,
            "when": """ansible_facts['os_family'] == 'Archlinux'""",
        },
    ]

    def setup(self):
        iis_shortname_jar = "https://github.com/irsdl/IIS-ShortName-Scanner/raw/master/iis_shortname_scanner.jar"
        iis_shortname_config = "https://raw.githubusercontent.com/irsdl/IIS-ShortName-Scanner/master/config.xml"
        self.iis_scanner_jar = self.helpers.download(iis_shortname_jar, cache_hrs=720)
        self.iis_scanner_config = self.helpers.download(iis_shortname_config, cache_hrs=720)
        if self.iis_scanner_jar and self.iis_scanner_config:
            return True
        return False

    def handle_event(self, event):

        normalized_url = event.data.rstrip("/") + "/"
        result = self.detect(normalized_url)

        if result:
            description = f"IIS Shortname Vulnerability"
            self.emit_event(
                {"severity": "LOW", "host": str(event.host), "url": normalized_url, "description": description},
                "VULNERABILITY",
                event,
            )
            if not self.config.get("detect_only"):
                command = [
                    "java",
                    "-jar",
                    self.iis_scanner_jar,
                    "0",
                    str(self.config.get("threads", 8)),
                    normalized_url,
                    self.iis_scanner_config,
                ]
                output = self.helpers.run(command).stdout
                self.debug(output)
                discovered_directories, discovered_files = self.shortname_parse(output)
                for d in discovered_directories:
                    if d[-2] == "~":
                        d = d.split("~")[:-1][0]
                    self.emit_event(normalized_url + d, "URL_HINT", event, tags=["directory"])
                for f in discovered_files:
                    if f[-2] == "~":
                        f = f.split("~")[:-1][0]
                    self.emit_event(normalized_url + f, "URL_HINT", event, tags=["file"])

    def detect(self, url):

        detected = False
        http_methods = ["GET", "OPTIONS", "DEBUG"]
        for http_method in http_methods:
            dir_name = self.helpers.rand_string(8)
            file_name = self.helpers.rand_string(1)
            control_url = url.rstrip("/") + "/" + f"{dir_name}*~1*/{file_name}.aspx"
            control = self.helpers.request(control_url, method=http_method)
            test_url = url.rstrip("/") + "/" + f"*~1*/{file_name}.aspx"
            test = self.helpers.request(test_url, method=http_method)
            if (control != None) and (test != None):
                if (control.status_code != 404) and (test.status_code == 404):
                    detected = True
        return detected

    def shortname_parse(self, output):
        discovered_directories = []
        discovered_files = []
        parseLines = output.split("\n")
        inDirectories = False
        inFiles = False
        for idx, line in enumerate(parseLines):
            if "Identified directories" in line:
                inDirectories = True
            elif "Indentified files" in line:
                inFiles = True
                inDirectories = False
            elif ":" in line:
                pass
            elif "Actual" in line:
                pass
            else:
                if inFiles == True:
                    if len(line) > 0:
                        shortname = line.split(" ")[-1].split(".")[0].split("~")[0]
                        extension = line.split(" ")[-1].split(".")[1]
                        if "?" not in extension:
                            discovered_files.append(f"{shortname}.{extension}".lower())

                elif inDirectories == True:
                    if len(line) > 0:
                        shortname = line.split(" ")[-1]
                        discovered_directories.append(shortname.lower())
        return discovered_directories, discovered_files



import ipaddress
import traceback

from bbot.modules.report.base import ReportModule


class asn(ReportModule):
    watched_events = ["IP_ADDRESS"]
    produced_events = ["ASN"]
    flags = ["passive", "subdomain-enum", "safe"]
    meta = {"description": "Query bgpview.io for ASNs"}
    scope_distance_modifier = 0

    base_url = "https://api.bgpview.io"

    def setup(self):
        self.asn_counts = {}
        self.asn_data = {}
        self.asn_metadata = {}
        return True

    def filter_event(self, event):
        if "private" in event.tags:
            return False
        return True

    def handle_event(self, event):
        if self.cache_get(event.host) == False:
            asns = self.get_asn(event.host)
            if asns:
                for asn in asns:
                    if asn not in self.asn_metadata:
                        contacts = self.get_asn_metadata(asn)
                        if not contacts:
                            continue
                        for c in contacts:
                            self.emit_event(c, "EMAIL_ADDRESS", source=event)
                        self.asn_metadata[asn] = True

    def report(self):
        asn_data = sorted(self.asn_data.items(), key=lambda x: self.asn_counts[x[0]], reverse=True)
        header = ["ASN", "Subnet", "Host Count", "Name", "Description"]
        table = []
        for subnet, prefix in asn_data:
            count = self.asn_counts[subnet]
            name = prefix.get("name", "")
            description = prefix.get("description", "")
            asn = "AS" + str(prefix.get("asn", {}).get("asn", ""))
            table.append([asn, subnet, f"{count:,}", name, description])
            event_str = f"{asn} - {subnet} ({count:,} hosts): {name}, {description}"
            self.emit_event(event_str, "ASN", source=self.scan.root_event, quick=True)
        for row in self.helpers.make_table(table, header).splitlines():
            self.info(row)

    def cache_get(self, ip):
        ret = False
        for p in self.helpers.ip_network_parents(ip):
            try:
                self.asn_counts[p] += 1
                if ret == False:
                    ret = p
            except KeyError:
                continue
        return ret

    def get_asn(self, ip):
        url = f"{self.base_url}/ip/{ip}"
        r = self.helpers.request(url, retries=5)
        try:
            j = r.json()
            data = j.get("data", {})
            if data:
                prefixes = data.get("prefixes", [])
                for prefix in prefixes:
                    subnet = prefix.get("prefix", "")
                    if not subnet:
                        continue
                    subnet = ipaddress.ip_network(subnet)
                    self.asn_data[subnet] = prefix
                    self.asn_counts[subnet] = 1
                    asn = str(prefix.get("asn", {}).get("asn", ""))
                    yield asn
            else:
                self.debug(f'No results for "{ip}"')
        except Exception as e:
            self.warning(f"Error retrieving ASN for {ip}: {e}")
            self.debug(f"Got data: {getattr(r, 'content', '')}")
            self.debug(traceback.format_exc())

    def get_asn_metadata(self, asn):
        url = f"{self.base_url}/asn/{asn}"
        r = self.helpers.request(url, retries=5)
        try:
            j = r.json()
            data = j.get("data", {})
            if not data:
                self.debug(f'No results for "{asn}"')
                return
            email_contacts = data.get("email_contacts", [])
            abuse_contacts = data.get("abuse_contacts", [])
            contacts = [l.strip().lower() for l in email_contacts + abuse_contacts]
            return list(set(contacts))
        except Exception as e:
            self.warning(f"Error retrieving ASN metadata for {asn}: {e}")
            self.debug(traceback.format_exc())



import queue
import logging
import threading
from time import sleep
from contextlib import suppress
from datetime import datetime, timedelta

from ..core.errors import ScanCancelledError, ValidationError

log = logging.getLogger("bbot.scanner.manager")


class ScanManager:
    """
    Manages modules and events during a scan
    """

    def __init__(self, scan):
        self.scan = scan
        self.event_queue = queue.PriorityQueue()
        self.events_distributed = set()
        self.events_distributed_lock = threading.Lock()
        self.events_accepted = set()
        self.events_accepted_lock = threading.Lock()
        self.events_resolved = dict()
        self.events_resolved_lock = threading.Lock()
        self.dns_resolution = self.scan.config.get("dns_resolution", False)

    def init_events(self):
        """
        seed scanner with target events
        """
        self.queue_event(self.scan.root_event)
        for event in self.scan.target.events:
            self.scan.verbose(f"Target: {event}")
            self.emit_event(event)
        # force submit batches
        for mod in self.scan.modules.values():
            mod._handle_batch(force=True)

    def emit_event(self, event, *args, **kwargs):
        quick = kwargs.pop("quick", False)
        release = kwargs.get("release", True)
        if quick:
            try:
                kwargs.pop("abort_if")
                kwargs.pop("on_success_callback")
                self.queue_event(event, *args, **kwargs)
            finally:
                if release:
                    with suppress(Exception):
                        event.module._event_semaphore.release()
        else:
            # don't raise an exception if the thread pool has been shutdown
            with suppress(RuntimeError):
                self.scan._event_thread_pool.submit_task(self.catch, self._emit_event, event, *args, **kwargs)

    def _emit_event(self, event, *args, **kwargs):
        release = kwargs.pop("release", True)
        emit_event = True
        try:
            on_success_callback = kwargs.pop("on_success_callback", None)
            abort_if = kwargs.pop("abort_if", None)
            log.debug(f'module "{event.module}" raised {event}')

            if event._dummy:
                log.warning(f"Cannot emit dummy event: {event}")
                return

            if event == event.get_source():
                log.debug(f"Omitting event with self as source: {event}")
                return

            # DNS resolution
            dns_children, dns_tags, event_whitelisted_dns, event_blacklisted_dns = self.scan.helpers.dns.resolve_event(
                event
            )
            event_whitelisted = event_whitelisted_dns | self.scan.whitelisted(event)
            event_blacklisted = event_blacklisted_dns | self.scan.blacklisted(event)
            if event.type in ("DNS_NAME", "IP_ADDRESS"):
                event.tags.update(dns_tags)
            if event_blacklisted:
                event.tags.add("blacklisted")

            # Blacklist purging
            if "blacklisted" in event.tags:
                reason = "event host"
                if event_blacklisted_dns:
                    reason = "DNS associations"
                log.debug(f"Omitting due to blacklisted {reason}: {event}")
                emit_event = False

            # Wait for parent event to resolve (in case its scope distance changes)
            while 1:
                if self.scan.stopping:
                    raise ScanCancelledError()
                resolved = event._resolved.wait(timeout=0.1)
                if resolved:
                    # update event's scope distance based on its parent
                    event.scope_distance = event.source.scope_distance + 1
                    break

            # Scope shepherding
            event_is_duplicate = self.is_duplicate_event(event)
            event_in_report_distance = event.scope_distance <= self.scan.scope_report_distance
            set_scope_distance = event.scope_distance
            if event_whitelisted:
                set_scope_distance = 0
            if event.host:
                if (event_whitelisted or event_in_report_distance) and not event_is_duplicate:
                    if set_scope_distance == 0:
                        log.debug(f"Making {event} in-scope")
                    event.make_in_scope(set_scope_distance)
                else:
                    if event.scope_distance > self.scan.scope_report_distance:
                        log.debug(
                            f"Making {event} internal because its scope_distance ({event.scope_distance}) > scope_report_distance ({self.scan.scope_report_distance})"
                        )
                        event.make_internal()
            else:
                log.debug(f"Making {event} in-scope because it does not have identifying scope information")
                event.make_in_scope(0)

            # now that the event is properly tagged, we can finally make decisions about it
            if callable(abort_if) and abort_if(event):
                log.debug(f"{event.module}: not raising event {event} due to custom criteria in abort_if()")
                return

            if not self.accept_event(event):
                return

            # queue the event before emitting its DNS children
            if emit_event:
                self.queue_event(event)

            if callable(on_success_callback):
                self.catch(on_success_callback, event)

            ### Emit DNS children ###
            emit_children = -1 < event.scope_distance < self.scan.dns_search_distance
            # speculate DNS_NAMES and IP_ADDRESSes from other event types
            source_event = event
            if event.host and event.type not in ("DNS_NAME", "IP_ADDRESS", "IP_RANGE"):
                source_module = self.scan.helpers._make_dummy_module("host", _type="internal")
                source_event = self.scan.make_event(event.host, "DNS_NAME", module=source_module, source=event)
                source_event.scope_distance = event.scope_distance
                if "target" in event.tags:
                    source_event.tags.add("target")
                if not str(event.module) == "speculate":
                    self.emit_event(source_event)
            if self.dns_resolution and emit_children:
                dns_child_events = []
                if dns_children:
                    for record, rdtype in dns_children:
                        module = self.scan.helpers.dns._get_dummy_module(rdtype)
                        try:
                            child_event = self.scan.make_event(record, "DNS_NAME", module=module, source=source_event)
                            dns_child_events.append(child_event)
                        except ValidationError as e:
                            log.warning(
                                f'Event validation failed for DNS child of {source_event}: "{record}" ({rdtype}): {e}'
                            )
                for child_event in dns_child_events:
                    self.emit_event(child_event)

        except ValidationError as e:
            log.warning(f"Event validation failed with args={args}, kwargs={kwargs}: {e}")
            import traceback

            log.debug(traceback.format_exc())

        finally:
            if release:
                with suppress(Exception):
                    event.module._event_semaphore.release()
                if emit_event:
                    self.scan.stats.event_emitted(event)

    def hash_event(self, event):
        """
        Hash an event for duplicate detection

        This is necessary because duplicate events from certain sources (e.g. DNS)
            need to be allowed in order to preserve their relationship trail
        """
        module_type = getattr(event.module, "_type", "")
        if module_type == "DNS":
            # allow duplicate events from dns resolution as long as their source event is unique
            return hash((event, str(event.module), event.source_id))
        else:
            return hash((event, str(event.module)))

    def is_duplicate_event(self, event, add=False):
        event_hash = self.hash_event(event)
        suppress_dupes = getattr(event.module, "suppress_dupes", True)
        with self.events_accepted_lock:
            duplicate_event = suppress_dupes and event_hash in self.events_accepted
            if add:
                self.events_accepted.add(event_hash)
        return duplicate_event and not event._force_output

    def accept_event(self, event):
        if self.is_duplicate_event(event, add=True):
            log.debug(f"{event.module}: not raising duplicate event {event}")
            return False
        return True

    def catch(self, callback, *args, **kwargs):
        """
        Wrapper to ensure error messages get surfaced to the user
        """
        ret = None
        on_finish_callback = kwargs.pop("_on_finish_callback", None)
        force = kwargs.pop("_force", False)
        try:
            if not self.scan.stopping or force:
                ret = callback(*args, **kwargs)
        except ScanCancelledError as e:
            log.debug(f"ScanCancelledError in {callback.__qualname__}(): {e}")
        except BrokenPipeError as e:
            log.debug(f"BrokenPipeError in {callback.__qualname__}(): {e}")
        except Exception as e:
            import traceback

            log.error(f"Error in {callback.__qualname__}(): {e}")
            log.debug(traceback.format_exc())
        except KeyboardInterrupt:
            log.debug(f"Interrupted")
            self.scan.stop()
        if callable(on_finish_callback):
            try:
                on_finish_callback()
            except Exception as e:
                import traceback

                log.error(
                    f"Error in on_finish_callback {on_finish_callback.__qualname__}() after {callback.__qualname__}(): {e}"
                )
                log.debug(traceback.format_exc())
        return ret

    def queue_event(self, *args, **kwargs):
        """
        Queue event with manager
        """
        event = self.scan.make_event(*args, **kwargs)
        self.event_queue.put(event)

    def distribute_event(self, event):
        """
        Queue event with modules
        """
        # TODO: save memory by removing reference to source object (this causes bugs)
        # if not event._internal:
        #    event._source = None

        dup = False
        event_hash = hash(event)
        # with self.events_distributed_lock:
        if event_hash in self.events_distributed:
            self.scan.verbose(f"{event.module}: Duplicate event: {event}")
            dup = True
        else:
            self.events_distributed.add(event_hash)
        # absorb event into the word cloud if it's in scope
        if not dup and -1 < event.scope_distance < 1:
            self.scan.word_cloud.absorb_event(event)
        stats_recorded = False
        for mod in self.scan.modules.values():
            if not dup or mod.accept_dupes:
                event_within_scope_distance = -1 < event.scope_distance <= self.scan.scope_search_distance
                event_within_report_distance = -1 < event.scope_distance <= self.scan.scope_report_distance
                if mod._type == "output":
                    if event_within_report_distance or (event._force_output and mod.emit_graph_trail):
                        mod.queue_event(event)
                        if not stats_recorded:
                            stats_recorded = True
                            self.scan.stats.event_produced(event)
                else:
                    if event_within_scope_distance:
                        mod.queue_event(event)

    def loop_until_finished(self):

        counter = 0
        event_counter = 0
        timedelta_2secs = timedelta(seconds=2)
        last_log_time = datetime.now()

        reported = False
        try:
            self.scan.dispatcher.on_start(self.scan)

            # watch for newly-generated events
            while 1:

                if self.scan.status == "ABORTING":
                    while 1:
                        try:
                            # Empty event queue
                            self.event_queue.get_nowait()
                        except queue.Empty:
                            break
                    break

                # print status every 2 seconds
                now = datetime.now()
                time_since_last_log = now - last_log_time
                if time_since_last_log > timedelta_2secs:
                    self.modules_status(_log=True, passes=1)
                    last_log_time = now

                try:
                    event = self.event_queue.get_nowait()
                    event_counter += 1
                except queue.Empty:
                    finished = self.modules_status().get("finished", False)
                    if finished and reported:
                        break
                    # If the scan finished
                    if finished:
                        # If new events were generated in the last iteration
                        if event_counter > 0:
                            self.scan.status = "FINISHING"
                            # Trigger .finished() on every module and start over
                            for mod in self.scan.modules.values():
                                mod.queue_event("FINISHED")
                            event_counter = 0
                        elif not reported:
                            # Run .report() on every module and start over
                            for mod in self.scan.modules.values():
                                self.catch(mod.report)
                            reported = True
                        else:
                            # Otherwise stop the scan if no new events were generated in this iteration
                            break
                    else:
                        # save on CPU
                        sleep(0.01)
                    counter += 1
                    continue

                # distribute event to modules
                self.distribute_event(event)

        except KeyboardInterrupt:
            self.scan.stop()

        except Exception:
            import traceback

            log.critical(traceback.format_exc())

    def modules_status(self, _log=False, passes=None):

        # If scan looks to be finished, check an additional five times to ensure that it really is
        # There is a tiny chance of a race condition, which this helps to avoid
        if passes is None:
            passes = 5
        else:
            passes = max(1, int(passes))

        finished = True
        while passes > 0:

            status = {"modules": {}, "scan": self.scan.status_detailed}

            for num_events in status["scan"]["queued_events"].values():
                if num_events > 0:
                    finished = False

            for num_tasks in status["scan"]["queued_tasks"].values():
                if num_tasks > 0:
                    finished = False

            for m in self.scan.modules.values():
                mod_status = m.status
                if mod_status["running"]:
                    finished = False
                status["modules"][m.name] = mod_status

            for mod in self.scan.modules.values():
                if mod.errored and mod.event_queue not in [None, False]:
                    with suppress(Exception):
                        mod.set_error_state()

            passes -= 1
            if finished and passes > 0:
                sleep(0.1)
            else:
                break

        status["finished"] = finished

        modules_running = [m for m, s in status["modules"].items() if s["running"]]
        modules_errored = [m for m, s in status["modules"].items() if s["errored"]]

        if _log:
            events_queued = [
                (m, (s["events"]["incoming"], s["events"]["outgoing"])) for m, s in status["modules"].items()
            ]
            events_queued.sort(key=lambda x: sum(x[-1]), reverse=True)
            events_queued = [(m, q) for m, q in events_queued if sum(q) > 0][:5]
            events_queued_str = ""
            if events_queued:
                events_queued_str = " (" + ", ".join([f"{m}: I:{i:,} O:{o:,}" for m, (i, o) in events_queued]) + ")"
            tasks_queued = [(m, s["tasks"]["total"]) for m, s in status["modules"].items()]
            tasks_queued.sort(key=lambda x: x[-1], reverse=True)
            tasks_queued = [(m, q) for m, q in tasks_queued if q > 0][:5]
            tasks_queued_str = ""
            if tasks_queued:
                tasks_queued_str = " (" + ", ".join([f"{m}: {q:,}" for m, q in tasks_queued]) + ")"

            num_events_queued = sum([sum(m[-1]) for m in events_queued])
            self.scan.hugeverbose(f"Events queued: {num_events_queued:,}{events_queued_str}")

            num_tasks_queued = sum([m[-1] for m in tasks_queued])
            self.scan.hugeverbose(f"Module tasks queued: {num_tasks_queued:,}{tasks_queued_str}")

            num_scan_tasks = status["scan"]["queued_tasks"]["total"]
            dns_tasks = status["scan"]["queued_tasks"]["dns"]
            event_tasks = status["scan"]["queued_tasks"]["event"]
            main_tasks = status["scan"]["queued_tasks"]["main"]
            internal_tasks = status["scan"]["queued_tasks"]["internal"]
            manager_events_queued = status["scan"]["queued_events"]["manager"]
            self.scan.hugeverbose(
                f"Scan tasks queued: {num_scan_tasks:,} (Main: {main_tasks:,}, Events: {event_tasks:,} waiting, {manager_events_queued:,} in queue, DNS: {dns_tasks:,}, Internal: {internal_tasks:,})"
            )

            if modules_running:
                self.scan.hugeverbose(
                    f'Modules running: {len(modules_running):,} ({", ".join([m for m in modules_running])})'
                )
            if modules_errored:
                self.scan.hugeverbose(
                    f'Modules errored: {len(modules_errored):,} ({", ".join([m for m in modules_errored])})'
                )

        status.update({"modules_running": len(modules_running), "modules_errored": len(modules_errored)})

        return status



# based on https://github.com/ElSicarius/interactsh-python/blob/main/sources/interactsh.py
import json
import base64
import random
import logging
from time import sleep
from uuid import uuid4
from threading import Thread
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Cipher import AES, PKCS1_OAEP

from bbot.core.errors import InteractshError

log = logging.getLogger("bbot.core.helpers.interactsh")

server_list = ["oast.pro", "oast.live", "oast.site", "oast.online", "oast.fun", "oast.me"]


class Interactsh:
    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.server = self.parent_helper.config.get("interactsh_server", None)
        self.token = self.parent_helper.config.get("interactsh_token", None)
        self._thread = None

    def register(self, callback=None):
        if self.server == None:
            self.server = random.choice(server_list)

        rsa = RSA.generate(1024)

        self.public_key = rsa.publickey().exportKey()
        self.private_key = rsa.exportKey()

        encoded_public_key = base64.b64encode(self.public_key).decode("utf8")

        uuid = uuid4().hex.ljust(33, "a")
        guid = "".join(i if i.isdigit() else chr(ord(i) + random.randint(0, 20)) for i in uuid)

        self.domain = f"{guid}.{self.server}"

        self.correlation_id = guid[:20]
        self.secret = str(uuid4())
        headers = {}

        if self.token:
            headers["Authorization"] = self.token

        data = {"public-key": encoded_public_key, "secret-key": self.secret, "correlation-id": self.correlation_id}
        r = self.parent_helper.request(
            f"https://{self.server}/register", headers=headers, json=data, method="POST", retries="infinite"
        )
        msg = r.json().get("message", "")
        if msg != "registration successful":
            raise InteractshError(f"Failed to register with interactsh server {self.server}")

        log.info(
            f"Successfully registered to interactsh server {self.server} with correlation_id {self.correlation_id} [{self.domain}]"
        )

        if callable(callback):
            self._thread = Thread(target=self.poll_loop, args=(callback,), daemon=True)
            self._thread.start()

        return self.domain

    def deregister(self):

        headers = {}
        if self.token:
            headers["Authorization"] = self.token

        data = {"secret-key": self.secret, "correlation-id": self.correlation_id}

        r = self.parent_helper.request(f"https://{self.server}/deregister", headers=headers, json=data, method="POST")
        if "success" not in r.text:
            raise InteractshError(f"Failed to de-register with interactsh server {self.server}")

    def poll(self):

        headers = {}
        if self.token:
            headers["Authorization"] = self.token

        r = self.parent_helper.request(
            f"https://{self.server}/poll?id={self.correlation_id}&secret={self.secret}", headers=headers
        )

        data_list = r.json().get("data", None)
        if data_list:
            aes_key = r.json()["aes_key"]

            for data in data_list:

                decrypted_data = self.decrypt(aes_key, data)
                yield decrypted_data

    def poll_loop(self, callback):
        return self.parent_helper.scan.manager.catch(self._poll_loop, callback, _force=True)

    def _poll_loop(self, callback):
        while 1:
            if self.parent_helper.scan.stopping:
                sleep(1)
                continue
            data_list = list(self.poll())
            if not data_list:
                sleep(10)
                continue
            for data in data_list:
                if data:
                    callback(data)

    def decrypt(self, aes_key, data):
        private_key = RSA.importKey(self.private_key)
        cipher = PKCS1_OAEP.new(private_key, hashAlgo=SHA256)
        aes_plain_key = cipher.decrypt(base64.b64decode(aes_key))
        decode = base64.b64decode(data)
        bs = AES.block_size
        iv = decode[:bs]
        cryptor = AES.new(key=aes_plain_key, mode=AES.MODE_CFB, IV=iv, segment_size=128)
        plain_text = cryptor.decrypt(decode)
        return json.loads(plain_text[16:])



from bbot.modules.base import BaseModule
import re


class aspnet_viewstate(BaseModule):

    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["VULNERABILITY"]
    flags = ["active", "safe", "web"]
    meta = {"description": "Parse web pages for viewstates and check them against blacklist3r"}

    generator_regex = re.compile(r'<input.+__VIEWSTATEGENERATOR"\svalue="(\w+)"')
    viewstate_regex = re.compile(r'<input.+__VIEWSTATE"\svalue="([^"]+)"')

    deps_ansible = [
        {
            "name": "apt_key",
            "become": True,
            "apt_key": {
                "keyserver": "hkp://keyserver.ubuntu.com:80",
                "id": "3FA7E0328081BFF6A14DA29AA6A19B38D3D831EF",
            },
            "when": """ansible_facts['os_family'] == 'Debian'""",
        },
        {
            "name": "Add Mono Repo",
            "become": True,
            "apt_repository": {"repo": "deb https://download.mono-project.com/repo/ubuntu stable-focal main"},
            "when": """ansible_facts['os_family'] == 'Debian'""",
        },
        {
            "name": "Install mono-devel (Debian, RedHat)",
            "become": True,
            "package": {"name": ["mono-devel", "zip"], "state": "latest"},
            "when": """ansible_facts['os_family'] in ['Debian', 'RedHat']""",
        },
        {
            "name": "Install mono (Archlinux)",
            "become": True,
            "package": {"name": "mono", "state": "latest"},
            "when": """ansible_facts['os_family'] == 'Archlinux'""",
        },
        {"name": "Create blacklist3r dir", "file": {"state": "directory", "path": "{BBOT_TOOLS}/blacklist3r/"}},
        {
            "name": "Unarchive blacklist3r",
            "unarchive": {
                "src": "https://github.com/NotSoSecure/Blacklist3r/releases/download/4.0/AspDotNetWrapper.zip",
                "include": ["MachineKeys.txt", "CommandLine.dll", "AspDotNetWrapper.exe"],
                "dest": "{BBOT_TOOLS}/blacklist3r/",
                "remote_src": True,
            },
        },
    ]

    def handle_event(self, event):

        generator_match = self.generator_regex.search(event.data.get("response-body", ""))
        viewstate_match = self.viewstate_regex.search(event.data.get("response-body", ""))

        if generator_match and viewstate_match:
            generator = generator_match.group(1)
            viewstate = viewstate_match.group(1)
            self.debug(f"Discovered viewstate for URL {event.data['url']}")
            self.emit_event(
                {"technology": "asp", "host": str(event.host), "url": event.data["url"]}, "TECHNOLOGY", event
            )
            self.emit_event(
                {"technology": "iis", "host": str(event.host), "url": event.data["url"]}, "TECHNOLOGY", event
            )
            tool_path = self.scan.helpers.tools_dir / "blacklist3r"
            command = [
                "mono",
                tool_path / "AspDotNetWrapper.exe",
                "--keypath",
                tool_path / "MachineKeys.txt",
                "--encrypteddata",
                f"{viewstate}",
                "--purpose=viewstate",
                f"--modifier={generator}",
                "--macdecode",
            ]
            output = self.helpers.run(command).stdout
            self.debug(f"blacklist3r output: {output}")
            if "Keys found!!" in output:
                for x in output.splitlines():
                    if "DecryptionKey" in x:
                        solvedDecryption = x.split(":")[1]
                    if "ValidationKey" in x:
                        solvedValidation = x.split(":")[1]

                description = (
                    f"Known MachineKey found. EncryptionKey: [{solvedDecryption}] ValidationKey: [{solvedValidation}]"
                )
                event_data = {
                    "severity": "CRITICAL",
                    "description": description,
                    "host": str(event.host),
                    "url": event.data["url"],
                }
                self.emit_event(event_data, "VULNERABILITY", event)
        else:
            self.debug("aspnet_viewstate viewstate not found")
